# Corner recovery
import svcco
import pyvista as pv
import matplotlib.pyplot as plt
import numpy as np
import matplotlib as mpl
cube = pv.Cube().triangulate().subdivide(3)

s = svcco.surface()
s.set_data(cube.points,cube.point_normals)
s.solve()
s.build()

def test(mpu_object,h=0,q_range=(1,8),resolution=20,k=2,buf=1.25,level=0,workers=1,plane_axis=2,plane_value=0.5,unit_scale='mm',name="untitled"):
    if plane_axis == 0:
        plane_value = (mpu_object.x_range[1]-mpu_object.x_range[0])*plane_value + mpu_object.x_range[0]
    elif plane_axis == 1:
        plane_value = (mpu_object.y_range[1]-mpu_object.y_range[0])*plane_value + mpu_object.y_range[0]
    elif plane_axis == 2:
        plane_value = (mpu_object.z_range[1]-mpu_object.z_range[0])*plane_value + mpu_object.z_range[0]
    colormap = plt.cm.RdBu
    #main_axes = list(range(len(DIMS)))
    #main_axes.remove(plane_axis)
    c = np.arange(q_range[0], q_range[1] + 1)
    norm = mpl.colors.Normalize(vmin=c.min(), vmax=c.max())
    cmap = mpl.cm.ScalarMappable(norm=norm, cmap=mpl.cm.Blues)
    cmap.set_array([])
    fig,ax = plt.subplots()
    for q in range(q_range[0],q_range[1]+1):
        mpu_object.build(q=q,h=h)
        DIMS,results_exact = svcco.implicit.visualize.visualize.mpu_meshgrid(mpu_object,res=resolution,workers=workers,
                                                                             k=len(mpu_object.patches),plane_axis=plane_axis,
                                                                             plane_value=plane_value,buf=buf)
        main_axes = list(range(len(DIMS)))
        main_axes.remove(plane_axis)
        CS = ax.contour(DIMS[main_axes[0]],DIMS[main_axes[1]],results_exact,levels=[level],linewidths=(2,))
        C = CS.collections[0]
        C.set_color(cmap.to_rgba(q+1))
        #fmt = {CS.levels[0]:r"$\Gamma_{q=" +str(i) + r"}$"}
        #ax.clabel(CS,fmt=fmt,fontsize=10)
    fig.colorbar(cmap, ticks=c,label=r"Patch Degree, $q$")
    name = "corner_recovery_q{}-{}.svg".format(q_range[0],q_range[1])
    fig.savefig(name,format="svg")
    plt.show()



contour="""contour_polydata = []
contour_list = []
contour_set_{} = []
contour_polydata_list_{} = []
contour_set_{}.append(segmentation.Circle({},center={},normal={}))
contour_polydata_list_{}.append(contour_set_{}[-1].get_polydata())
contour_list.append(contour_set_{})
contour_polydata.append(contour_polydata_list_{})
if {}:
    dmg.add_segmentation('vessel_{}','vessel_{}',contour_set_{})
"""



import numba as nb
import numpy as np

@nb.jit(nopython=True,cache=True,nogil=True)
def radii(data,Pperm,Pterm):
    updated_radii = [0]
    data[0,21] = ((data[0,25]*data[0,22])/(Pperm-Pterm)) ** (1/4)
    if data[0,15] < 0:
        return
    children_d = data[0,15:17].flatten()
    children_i = children_d.astype(np.int64)
    parents_d = data[children_i,17].flatten()
    parents_i = parents_d.astype(np.int64)
    bifs = data[parents_i[::2],23:25].flatten()
    radii = data[parents_i,21].flatten()
    while len(children_i) > 0:
        data[children_i,21] = bifs*radii
        updated_radii.extend(children_i)
        next_children = np.argwhere(data[children_i,15] > 0)
        next_children = children_i[next_children.flatten()]
        children_d = data[next_children,15:17].flatten()
        children_i = children_d.astype(np.int64)
        parents_d = data[children_i,17].flatten()
        parents_i = parents_d.astype(np.int64)
        bifs = data[parents_i[::2],23:25].flatten()
        radii = data[parents_i,21].flatten()
    return updated_radii

def update_radii(data,Pperm,Pterm):
    data[0,21] = ((data[0,25]*data[0,22])/(Pperm-Pterm)) ** (1/4)
    idx = [int(data[0,15]),int(data[0,16])]
    while len(idx) > 0:
        tmp = []
        for vessel in idx:
            parent = int(data[vessel,17])
            LR  = data[parent,15:17].astype(int)
            BIF = data[parent,23:25].flatten()
            LR_type = np.argwhere(LR == vessel).flatten()[0]
            BIF = BIF[LR_type]
            data[vessel,21] = data[parent,21]*BIF
            data[vessel,28] = data[parent,28]*BIF
            if data[vessel,15]>0:
                tmp.append(int(data[vessel,15]))
            if data[vessel,16]>0:
                tmp.append(int(data[vessel,16]))
        idx = tmp


#def naive_radii(data,Pperm,Pterm,idx):
#    if idx == 0:
#        data[idx,21] = ((data[idx,25]*data[idx,22])/(Pperm-Pterm)) ** (1/4)
#    elif data[idx,15] > 0 and data[idx,16] > 0:

##################################################################
# UNIT TEST
##################################################################



import numpy as np
#import pygmo as pg
from scipy import optimize
from functools import partial
from .solver_functions.init_normals_given import init_normals_given
from .solver_functions.init_normals_not_given import init_normals_not_given
#import sys #should remove

#sys.path.append('..') # should remove
from ..kernel.kernel import kernel

class solver:
    def __init__(self,points,normals):
        self.points  = points
        self.normals = normals
        self.kernel  = kernel(points)
        #self.problem = pg.problem(self.kernel)

    def set_solver(self,verbose):
        #nlopt_solvers = ['cobyla','bobyqa','newuoa','newuoa','newuoa_bound',
        #                 'praxis','neldermead','sbplx','mma','ccsaq','slsqp',
        #                 'lbfgs','tnewton_precond_restart','tnewton_precond',
        #                 'tnewton','var2','var1','auglag','auglag_eq']
        scipy_solvers = ['Nelder-Mead','Powell','CG','BFGS','Newton-CG',
                         'L-BFGS-B','TNC','COBYLA','SLSQP','trust-constr',
                         'dogleg','trust-ncg','trust-exact','trust-krylov']
        #if self.method in nlopt_solvers:
        #    nl = pg.nlopt(self.method)
        #    nl.xtol_rel = 1E-7
        #    nl.ftol_rel = 1E-7
        #    if self.method == 'lbfgs':
        #        nl.maxeval = 3000
        #    algorithm = pg.algorithm(nl)
        #    if verbose:
        #        algorithm.set_verbosity(1)
        if self.method in scipy_solvers:
            if verbose:
                if self.method == 'trust-constr':
                    options = {'disp':True}
                elif self.method == 'L-BFGS-B':
                    options = {'disp':99}
                elif self.method in ['Newton-CG','CG','BFGS','Nelder-Mead',
                                     'Powell','TNC','COBYLA','SLSQP']:
                    options = {'disp':True}
                else:
                    print('No verbosity allowed for this method.')
                    options = {}
            else:
                options = {}
            algorithm = partial(optimize.minimize,method=self.method,
                                    tol=1e-07,options=options)
            #algorithm = pg.algorithm(scp)
        else:
            print('Not an available solver method.')
            #print('See nlopt methods: {}'.format(nlopt_solvers))
            print('See scipy methods: {}'.format(scipy_solvers))
        self.algorithm = algorithm

    def vector_solver(self):
        #self.population = pg.population(self.problem)
        initial_solution = init_normals_given(self.normals)
        #self.population.push_back(initial_solution)
        #results = self.algorithm.evolve(self.population)
        bounds = []
        kernel_bounds = self.kernel.get_bounds()
        for i in range(len(kernel_bounds[0])):
            bounds.append(tuple([kernel_bounds[0][i],kernel_bounds[1][i]]))
        results = self.algorithm(self.kernel.fitness,x0=initial_solution,
                                 jac=self.kernel.jac,bounds=bounds)
        return results

    def variational_solver(self,lam):
        #self.population = pg.population(self.problem)
        initial_solution = init_normals_not_given(lam,self.kernel.K00,
                                                      self.kernel.K01,
                                                      self.kernel.K11)
        #self.population.push_back(initial_solution)
        #results = self.algorithm.evolve(self.population)
        bounds = []
        kernel_bounds = self.kernel.get_bounds()
        for i in range(len(kernel_bounds)):
            bounds.append(tuple([kernel_bounds[0][i],kernel_bounds[1][i]]))
        results = self.algorithm(self.kernel.fitness,x0=initial_solution,
                                 jac=self.kernel.jac,bounds=bounds)
        return results

    def solve(self,seed_number=1,lb=0.01,ub=1,perturb=0.01,
               local_verbosity=False,local_method='L-BFGS-B',
               variational=False,solver_method='Bounded',
               solver_verbosity=True):
        self.method = local_method
        self.set_solver(local_verbosity)
        if variational:
            fit = lambda x: self.variational_solver(x).champion_f
            lam = optimize.minimize_scalar(fit,bounds=(lb,ub),tol=1e-07,
                                           method=solver_method,options={'disp':3})
            result = self.variational_solver(lam.x)
        else:
            result = self.vector_solver()
        gg = result.x
        n = self.kernel.ndim
        g = np.ones(n*self.kernel.ddim)
        M_inv = self.kernel.A_inv[:n*(self.kernel.ddim+1),:n*(self.kernel.ddim+1)]
        N_inv = self.kernel.A_inv[:n*(self.kernel.ddim+1),n*(self.kernel.ddim+1):]
        ##########################################
        # Check normal direction
        ##########################################
        for i in range(n):
            g[i] = np.cos(gg[i*2+1])*np.sin(gg[i*2])
            g[i+n] = np.sin(gg[i*2+1])*np.sin(gg[i*2])
            g[i+2*n] = np.cos(gg[i*2])
        s = np.zeros(n)
        l_side = np.zeros(M_inv.shape[1])
        l_side[:len(s)] = s
        l_side[len(s):(len(s)+len(g))] = g
        a = np.matmul(M_inv,l_side)
        b = np.matmul(N_inv.T,l_side)
        return a,b



import numpy as np
import numba as nb
from scipy import interpolate

def get_local_points(data,vessel,terminal,num,isclamped):
    proximal = data[vessel,0:3]
    distal   = data[vessel,3:6]
    direction = distal - proximal
    direction = direction/np.linalg.norm(direction)
    #length = np.linalg.norm(direction)
    if np.all(vessel == 0) and isclamped:
        length = data[vessel,20]
        #print(length*np.linspace(0,1,num))
        shift = direction.reshape(-1,1)*length*np.linspace(0,1,num)
        points = proximal.reshape(1,-1) + shift.T
    else:
        line   = np.linspace(0,1,num)
        s,t    = np.meshgrid(line,line)
        s      = s.flatten().reshape(-1,1)
        t      = t.flatten().reshape(-1,1)
        points = proximal.reshape(1,-1)*(1-t)*s + \
                   distal.reshape(1,-1)*(t*s)+\
                   terminal.reshape(1,-1)*(1-s)
        #print('SHAPE: {}'.format(points.shape))
    return points

"""
@nb.jit(nopython=True,cache=True)
def angle_constraint(p0,p1,p2,constraint,greater):
    tmp = []
    for i in range(p0.shape[0]):
        vector_1 = p1 - p0[i]
        vector_2 = p2 - p0[i]
        vector_1 = vector_1 / np.linalg.norm(vector_1)
        vector_2 = vector_2 / np.linalg.norm(vector_2)
        if greater:
            if np.dot(vector_1,vector_2) > constraint:
                tmp.append(p0[i])
        else:
            if np.dot(vector_1,vector_2) < constraint:
                tmp.append(p0[i])
    return tmp
"""

def angle_constraint(p0,p1,p2,constraint,greater):
    vector_1 = p1 - p0
    vector_2 = p2 - p0
    vector_1 = vector_1 / np.linalg.norm(vector_1,axis=1).reshape(-1,1)
    vector_2 = vector_2 / np.linalg.norm(vector_2,axis=1).reshape(-1,1)
    angle    = np.array([np.dot(vector_1[i],vector_2[i]) for i in range(len(vector_1))])
    if greater:
        p0   = p0[angle>constraint,:]
    else:
        p0   = p0[angle<constraint,:]
    return p0


@nb.jit(nopython=True,cache=True)
def relative_length_constraint(p0,p1,p2,p3,ratio):
    d_max    = min([np.linalg.norm(p1-p2),
                    np.linalg.norm(p1-p3),
                    np.linalg.norm(p2-p3)])*ratio
    tmp = []
    for i in range(p0.shape[0]):
        if not np.linalg.norm(p1-p0[i])>d_max:
            continue
        if not np.linalg.norm(p2-p0[i])>d_max:
            continue
        if not np.linalg.norm(p3-p0[i])>d_max:
            continue
        tmp.append(p0[i])
    #p0 = np.array(tmp)
    #p0 = p0[np.linalg.norm(p1-p0,axis=1)>d_max]
    #p0 = p0[np.linalg.norm(p2-p0,axis=1)>d_max]
    #p0 = p0[np.linalg.norm(p3-p0,axis=1)>d_max]
    return tmp

def boundary_constraint(p0,boundary,patches):
    points = []
    for idx in range(p0.shape[0]):
        if boundary.within(p0[idx,0],p0[idx,1],p0[idx,2],patches):
            points.append(p0[idx,:])
    return np.array(points)

def absolute_length_constraint():
    pass

"""
class triangle:
    def __init__(self,level,ct,p0,p1,p2,p0_val,p1_val,degree=2,nodes=None,target=None):
        self.degree = degree
        if np.all(nodes) is not None:
            self.nodes = nodes
            self.targets = np.ones(self.nodes.shape[0])*target
        else:
            self.nodes,self.targets = self.sample(level,ct,degree,p0,p1,p2,p0_val,p1_val)

    def sample(self,level,ct,degree,p0,p1,p2,p0_val,p1_val):
        if degree == 1:
            nodes = []
            pts = np.array([p0,p1,p2])
            #for pt in pts:
            #    nodes.append(level(*pt,ct).x)
            nodes = pts
        elif degree == 2:
            nodes = []
            pts = np.array([p0,(p0+p1)/2,(p0+p2)/2,p1,(p1+p2)/2,p2])
            tar = np.array([p0_val,(p0_val+p1_val)/2,(p0_val+ct)/2,p1_val,(p1_val+ct)/2,ct])
            for idx,pt in enumerate(pts):
                if idx == 0 or idx==3 or idx == 5:
                    nodes.append(pt)
                else:
                    nodes.append(level(pt[0],pt[1],pt[2],tar[idx]).x)
        elif degree == 3:
            nodes = []
            pts = np.array([p0,(2*p0+p1)/3,(2*p0+p2)/3,(p0+2*p1)/3,(p0+p1+p2)/3,(p0+2*p2)/3,p1,(2*p1+p2)/3,(p1+2*p2)/3,p2])
            tar = np.array([p0_val,(2*p0_val+p1_val)/3,(2*p0_val+ct)/3,(p0_val+2*p1_val)/3,(p0_val+p1_val+ct)/3,(p0_val+2*ct)/3,p1_val,(2*p1_val+ct)/3,(p1_val+2*ct)/3,ct])
            for idx,pt in enumerate(pts):
                if idx == 0 or idx == 6 or idx == 9:
                    nodes.append(pt)
                else:
                    nodes.append(level(*pt,tar[idx]).x)
        nodes = np.array(nodes)
        return nodes,tar

    def cubic(self,s,t,u):
        coef = np.array([t**3,3*s*t**2,3*u*t**2,3*s**2*t,6*s*t*u,
                         3*t*u**2,s**3,3*s**2*u,3*s*u**2,u**3])
        #point = np.sum(coef*self.nodes,axis=0)
        point = coef[0,:].reshape(-1,1)*self.nodes[0,:].reshape(1,-1) +\
                coef[1,:].reshape(-1,1)*self.nodes[1,:].reshape(1,-1) +\
                coef[2,:].reshape(-1,1)*self.nodes[2,:].reshape(1,-1) +\
                coef[3,:].reshape(-1,1)*self.nodes[3,:].reshape(1,-1) +\
                coef[4,:].reshape(-1,1)*self.nodes[4,:].reshape(1,-1) +\
                coef[5,:].reshape(-1,1)*self.nodes[5,:].reshape(1,-1) +\
                coef[6,:].reshape(-1,1)*self.nodes[6,:].reshape(1,-1) +\
                coef[7,:].reshape(-1,1)*self.nodes[7,:].reshape(1,-1) +\
                coef[8,:].reshape(-1,1)*self.nodes[8,:].reshape(1,-1) +\
                coef[9,:].reshape(-1,1)*self.nodes[9,:].reshape(1,-1)
        target = coef[0,:].reshape(-1,1)*self.targets[0].reshape(1,-1) +\
                 coef[1,:].reshape(-1,1)*self.targets[1].reshape(1,-1) +\
                 coef[2,:].reshape(-1,1)*self.targets[2].reshape(1,-1) +\
                 coef[3,:].reshape(-1,1)*self.targets[3].reshape(1,-1) +\
                 coef[4,:].reshape(-1,1)*self.targets[4].reshape(1,-1) +\
                 coef[5,:].reshape(-1,1)*self.targets[5].reshape(1,-1) +\
                 coef[6,:].reshape(-1,1)*self.targets[6].reshape(1,-1) +\
                 coef[7,:].reshape(-1,1)*self.targets[7].reshape(1,-1) +\
                 coef[8,:].reshape(-1,1)*self.targets[8].reshape(1,-1) +\
                 coef[9,:].reshape(-1,1)*self.targets[9].reshape(1,-1)
        return point,target

    def quart(self,s,t,u):
        coef = np.array([t**2,2*s*t,2*t*u,s**2,2*s*u,u**2])
        #point = np.sum(coef*self.nodes,axis=0)
        point = coef[0,:].reshape(-1,1)*self.nodes[0,:].reshape(1,-1) +\
                coef[1,:].reshape(-1,1)*self.nodes[1,:].reshape(1,-1) +\
                coef[2,:].reshape(-1,1)*self.nodes[2,:].reshape(1,-1) +\
                coef[3,:].reshape(-1,1)*self.nodes[3,:].reshape(1,-1) +\
                coef[4,:].reshape(-1,1)*self.nodes[4,:].reshape(1,-1) +\
                coef[5,:].reshape(-1,1)*self.nodes[5,:].reshape(1,-1)
        target = coef[0,:].reshape(-1,1)*self.targets[0].reshape(1,-1) +\
                 coef[1,:].reshape(-1,1)*self.targets[1].reshape(1,-1) +\
                 coef[2,:].reshape(-1,1)*self.targets[2].reshape(1,-1) +\
                 coef[3,:].reshape(-1,1)*self.targets[3].reshape(1,-1) +\
                 coef[4,:].reshape(-1,1)*self.targets[4].reshape(1,-1) +\
                 coef[5,:].reshape(-1,1)*self.targets[5].reshape(1,-1)
        return point,target

    def linear(self,s,t,u):
        coef = np.array([t,s,u])
        #point = np.sum(coef*self.nodes,axis=0)
        point = coef[0,:].reshape(-1,1)*self.nodes[0,:].reshape(1,-1) +\
                coef[1,:].reshape(-1,1)*self.nodes[1,:].reshape(1,-1) +\
                coef[2,:].reshape(-1,1)*self.nodes[2,:].reshape(1,-1)
        target = coef[0,:].reshape(-1,1)*self.targets[0].reshape(1,-1) +\
                 coef[1,:].reshape(-1,1)*self.targets[1].reshape(1,-1) +\
                 coef[2,:].reshape(-1,1)*self.targets[2].reshape(1,-1)
        return point,target

    def generate(self,n=10):
        s = np.ones((n,n))*np.array([np.flip(np.linspace(0,1,num=n))]).T
        t = np.linspace(0,1-np.linspace(0,1,num=n),num=n)
        u = np.ones((n,n)) - s - t
        s = s[1:].flatten()
        t = t[1:].flatten()
        u = u[1:].flatten()
        if self.degree == 1:
            points,targets = self.linear(s,t,u)
        elif self.degree == 2:
            points,targets = self.quart(s,t,u)
        elif self.degree == 3:
            points,targets = self.cubic(s,t,u)
        else:
            points = None
            targets = None
        return points,targets,(s,t,u)

    def get_jit(self):
        if self.degree == 3:
            @nb.jit(nopython=True,cache=True)
            def cubic(s,t,u,nodes=self.nodes):
                coef = np.array([t**3,3*s*t**2,3*u*t**2,3*s**2*t,6*s*t*u,
                                 3*t*u**2,s**3,3*s**2*u,3*s*u**2,u**3])
                point = coef[0,:].reshape(-1,1)*nodes[0,:].reshape(1,-1) +\
                        coef[1,:].reshape(-1,1)*nodes[1,:].reshape(1,-1) +\
                        coef[2,:].reshape(-1,1)*nodes[2,:].reshape(1,-1) +\
                        coef[3,:].reshape(-1,1)*nodes[3,:].reshape(1,-1) +\
                        coef[4,:].reshape(-1,1)*nodes[4,:].reshape(1,-1) +\
                        coef[5,:].reshape(-1,1)*nodes[5,:].reshape(1,-1) +\
                        coef[6,:].reshape(-1,1)*nodes[6,:].reshape(1,-1) +\
                        coef[7,:].reshape(-1,1)*nodes[7,:].reshape(1,-1) +\
                        coef[8,:].reshape(-1,1)*nodes[8,:].reshape(1,-1) +\
                        coef[9,:].reshape(-1,1)*nodes[9,:].reshape(1,-1)
                return point
            return cubic

class b_tri:
    def __init__(self,points):
        #Check for the correct number of control points
        degree = (-3+(9-4*(2-2*points.shape[0]))**(1/2))/2
        degree_int = round(degree)
        self.points = points
        if np.isclose(degree,degree_int,0.0001):
            self.degree = degree_int
        else:
            print('Incorrect number of control points')
        if degree > 6:
            print('Number of control points greater than degrees.')
    def sample(self,n=10):
        s = np.ones((n,n))*np.array([np.flip(np.linspace(0,1,num=n))]).T
        t = np.linspace(0,1-np.linspace(0,1,num=n),num=n)
        u = np.ones((n,n)) - s - t
        s = s[1:].flatten()
        t = t[1:].flatten()
        u = u[1:].flatten()
        if self.degree == 1:
            coef = np.array([t,s,u]).T
            samples = []
            for i in range(coef.shape[0]):
                tmp = np.sum(self.points*coef[i,:],axis=0)
                samples.append(tmp)
        elif self.degree == 2:
            coef = np.array([t**2,2*s*t,2*t*u,s**2,2*s*u,u**2]).T
            samples = []
            for i in range(coef.shape[0]):
                tmp = np.sum(self.points*coef[i,:],axis=0)
                samples.append(tmp)
        elif self.degree == 3:
            coef = np.array([t**3,3*s*t**2,3*u*t**2,3*s**2*t,6*s*t*u,
                             3*t*u**2,s**3,3*s**2*u,3*s*u**2,u**3]).T
            samples = []
            for i in range(coef.shape[0]):
                tmp = np.sum(self.points*coef[i,:],axis=0)
                samples.append(tmp)
        elif self.degree == 4:
            coef = np.array([t**4,4*s*t**3,4*u*t**3,6*t**2*s**2,12*t**2*u,6*t**2*u**2,
                             4*t*s**3,12*t*s**2*u,12*t*s*u**2,s**4,4*s**3*u,6*s**2*u**2,
                             4*s*u**3,u**4]).T
            samples = []
            for i in range(coef.shape[0]):
                tmp = np.sum(self.points*coef[i,:],axis=0)
                samples.append(tmp)
        samples = np.array(samples)
        return samples
"""



import svcco
import numpy as np

pts = np.genfromtxt('D:\\Tree\\Tree_7-0\\svcco\\implicit\\tests\\heart_points_unique.csv',delimiter=',')
n =  np.genfromtxt('D:\\Tree\\Tree_7-0\\svcco\\implicit\\tests\\heart_normals_unique.csv',delimiter=',')

s = svcco.surface()
s.set_data(pts,normals=n)
s.solve()
s.build(q=2,resolution=40,k=20,buffer=2)



import svcco
import numpy as np
import matplotlib.pyplot as plt
import pyvista as pv
from tqdm import tqdm
import seaborn as sns

cube = pv.Cube(x_length=3.72,y_length=3.72,z_length=3.72).triangulate().subdivide(5)
sphere = pv.Sphere().triangulate().subdivide(1)

s = svcco.surface()
s.set_data(cube.points,cube.point_normals)
s.solve()

"""
heart = svcco.surface()
heart_points = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_points_unique.csv',delimiter=',')
heart_normals = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_normals_unique.csv',delimiter=',')
heart.set_data(heart_points,heart_normals)
heart.solve()

disk = pv.Disc(r_res=20,c_res=100)
cyl  = disk.extrude([0,0,1],capping=True).triangulate()
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl)
cyl  = cyl.subdivide(2)
cyl  = cyl.compute_normals(auto_orient_normals=True)
c = svcco.surface()
c.set_data(cyl.points,cyl.point_normals)
c.solve()


left_gyrus   = "D:\\Tree\\Tree_8-0\\brain_testing\\FJ3801_BP58201_FMA72658_Left inferior frontal gyrus.obj"

gyrus = svcco.surface()
gyrus.load(left_gyrus)
gyrus.solve()
"""
def test(surface_object,patch_size_range=(10,20),range_step=10):
    points = surface_object.points
    normals = surface_object.normals
    lower_size = []
    all_cond = []
    all_sizes = []
    start = patch_size_range[0]
    while start < patch_size_range[1]:
        surface_object.set_data(points,normals,local_min=start,local_max=2*start)
        surface_object.solve(quiet=False)
        condition_numbers = []
        patch_sizes = []
        for patch in surface_object.patches:
            condition_numbers.append(np.linalg.cond(patch.A_inv))
            patch_sizes.append(patch.points.shape[0])
        all_cond.append(np.array(condition_numbers))
        all_sizes.append(np.array(patch_sizes))
        lower_size.append(int(start))
        start += range_step
    lower_size = np.array(lower_size)
    #all_cond = np.array(all_cond)
    figure,ax = plt.subplots()
    ax.boxplot(all_cond,labels=lower_size)
    ax.set_yscale('log')
    plt.savefig('condition_number_plot.svg', format="svg", bbox_inches='tight')
    figure,ax = plt.subplots()
    ax.boxplot(all_sizes,labels=lower_size)
    plt.savefig('condition_patch_size_plot.svg', format="svg", bbox_inches='tight')



# Script for connecting trees of closed vascular networks
from scipy.spatial.distance import cdist
from scipy.optimize import linear_sum_assignment
import numpy as np

def connect(forest,network_id=-1,buffer=None):
    connections = []
    assignments = []
    if network_id == -1:
        for network in range(forest.number_of_networks):
            network_connections = []
            network_assignments = []
            tree_0 = forest.networks[network][0].data
            tree_1 = forest.networks[network][1].data
            terminals_0 = tree_0[np.argwhere(tree_0[:,15]==-1).flatten(),:]
            terminals_0 = terminals_0[np.argwhere(terminals_0[:,16]==-1).flatten(),:]
            terminals_0_ind = terminals_0[:,-1].astype(int).flatten()
            terminals_0_pts = terminals_0[:,3:6]
            terminals_1 = tree_1[np.argwhere(tree_1[:,15]==-1).flatten(),:]
            terminals_1 = terminals_1[np.argwhere(terminals_1[:,16]==-1).flatten(),:]
            terminals_1_ind = terminals_1[:,-1].astype(int).flatten()
            terminals_1_pts = terminals_1[:,3:6]
            network_assignments.append(terminals_0_ind)

            C = cdist(terminals_0_pts,terminals_1_pts)
            if not forest.convex:
                W = np.zeros(C.shape)
                penalty = np.max(C)
                for i in range(C.shape[0]):
                    for j in range(C.shape[1]):
                        mid = (terminals_0_pts[i] + terminals_1_pts[j])/2
                        mid = mid.flatten()
                        if not forest.boundary.within(mid[0],mid[1],mid[2],2):
                            W[i,j] = penalty
                C = C + W
            _,assignment = linear_sum_assignment(C)
            network_assignments.append(terminals_1_ind[assignment])
            midpoints  = (terminals_0_pts + terminals_1_pts[assignment])/2
            network_connections.append(midpoints)
            if forest.trees_per_network[network] > 2:
                for N in range(2,forest.trees_per_network[network]):
                    tree_n = forest.networks[network][N].data
                    terminals_n = tree_n[np.argwhere(tree_n[:,15]==-1).flatten(),:]
                    terminals_n = terminals_n[np.argwhere(terminals_n[:,16]==-1).flatten(),:]
                    terminals_n_ind = terminals_n[:,-1].astype(int).flatten()
                    terminals_n_pts = terminals_n[:,3:6]

                    C = cdist(midpoints,terminals_n_pts)
                    if not forest.convex:
                        W = np.zeros(c.shape)
                        penalty = np.max(C)
                        for i in range(C.shape[0]):
                            for j in range(C.shape[1]):
                                mid = (midpoints[i] + terminals_n_pts[j])/2
                                mid = mid.flatten()
                                if not forest.boundary.within(mid[0],mid[1],mid[2],2):
                                    W[i,j] = penalty
                        C = C + W
                    _,assignment = linear_sum_assignment(C)
                    network_assignments.append(terminals_n_ind[assignment])
            connections.append(network_connections)
            assignments.append(network_assignments)
    return connections, assignments

from copy import deepcopy
import pyvista as pv

def rotate(R0,R1,P0,P1,theta):
    theta = (theta/180)*np.pi
    Rvector = R1-R0
    Rvector = Rvector/np.linalg.norm(Rvector)
    vector = P1-P0
    vector = vector/np.linalg.norm(vector)
    vector_parallel = (np.dot(vector,Rvector.T)/np.dot(Rvector,Rvector.T))*Rvector
    vector_perp     = vector-vector_parallel
    w      = np.cross(Rvector,vector_perp)
    vector_perp_norm = np.linalg.norm(vector_perp)
    w_norm = np.linalg.norm(w)
    if vector_perp_norm > 0 and w_norm > 0:
        x1     = np.cos(theta)/np.linalg.norm(vector_perp)
        x2     = np.sin(theta)/np.linalg.norm(w)
        rotated_vector = np.linalg.norm(vector_perp)*(x1*vector_perp+x2*w)+vector_parallel
    else:
        #print(vector_perp)
        #print(w)
        rotated_vector = vector_parallel
    new_point = P0+rotated_vector*(np.linalg.norm(P1-P0))
    return new_point

def angle(P0,P1,P2):
    vector_1 = P0 - P1
    vector_2 = P2 - P1
    vector_1 = vector_1/np.linalg.norm(vector_1)
    vector_2 = vector_2/np.linalg.norm(vector_2)
    result   = (np.arccos(np.dot(vector_1,vector_2))/np.pi)*180
    return result


def parallel(P0,P1,P2):
    vector_1 = P0 - P1
    vector_2 = P2 - P1
    vector_1 = vector_1/np.linalg.norm(vector_1)
    vector_2 = vector_2/np.linalg.norm(vector_2)
    return np.all(np.isclose(vector_1,vector_2))

"""
def rotate_terminals(forest_copy):
    ANGLE_STEP = 5
    for ndx in range(forest_copy.number_of_networks):
        #for tdx in range(forest_copy.trees_per_network[ndx]):
        #A = forest_copy.assignments[ndx][tdx]
        #T = forest_copy.networks[ndx][tdx].data
        #C = forest_copy.connections[ndx][0]
        for edx in range(len(forest_copy.connections[ndx][0])):
            #CC = []
            #if T[A[edx],17] < 0:
            #    #Terminal root (cannot rotate)
            #    continue
            #elif parallel(T[A[edx],0:3],T[A[edx],3:6],TT[AA[edx],3:6]):
            #    continue
            #ANGLE = angle(T[A[edx],0:3],T[A[edx],3:6],TT[AA[edx],3:6])
            #attempt = 0
            #parent  = int(T[A[edx],17])
            #score  = [ANGLE]
            #center = (T[A[edx],3:6]+TT[AA[edx],3:6])/2
            #CC.append(center)
            #pts   = [deepcopy(T[A[edx],3:6])]
            max_attempt = int(360/ANGLE_STEP)
            CC = []
            C = forest_copy.connections[ndx][0]
            score  = [ANGLE]
            for kdx in range(forest_copy.trees_per_network[ndx]):
                A = forest_copy.assignments[ndx][kdx]
                T = forest_copy.networks[ndx][kdx].data
                #C = forest_copy.connections[ndx][0]
                ANGLE = angle(T[A[edx],0:3],T[A[edx],3:6],C[edx]])
                attempt = 0
                parent  = int(T[A[edx],17])
                score  = [ANGLE]
                CC.append(deepcopy(C[edx]))
                T = T = forest_copy.networks[ndx][tdx].data
                while attempt < max_attempt:
                    R0 = deepcopy(T[parent,0:3])
                    R1 = deepcopy(T[parent,3:6])
                    P0 = deepcopy(T[A[edx],0:3])
                    P1 = deepcopy(T[A[edx],3:6])
                    T[A[edx],3:6] = rotate(R0,R1,P0,P1,ANGLE_STEP)
                    T[A[edx],12:15] = (T[A[edx],3:6]-T[A[edx],0:3])/np.linalg.norm(T[A[edx],3:6]-T[A[edx],0:3])
                    center = np.zeros(3)
                    for tdx in range(forest_copy.trees_per_network[ndx]):
                        AA = forest_copy.assignments[ndx][tdx]
                        TT = forest_copy.networks[ndx][tdx].data
                        center += TT[AA[edx],3:6]
                    center = center/len(forest_copy.trees_per_network[ndx])
                    C[edx] = center
                    ANGLE = angle(T[A[edx],0:3],T[A[edx],3:6],C[edx]])
                    #center = (T[A[edx],3:6]+TT[AA[edx],3:6])/2
                    CC.append(deepcopy(C[edx]))
                    #view(forest_copy,highlight=[ndx,tdx,int(T[A[edx],-1])])
                    score.append(ANGLE)
                    pts.append(deepcopy(T[A[edx],3:6]))
                    attempt += 1
            if max(score) < 90:
                print("Fail")
                print(max(score))
            best = np.argmax(score)
            T[A[edx],3:6] = pts[best]
            T[A[edx],12:15] = (T[A[edx],3:6]-T[A[edx],0:3])/np.linalg.norm(T[A[edx],3:6]-T[A[edx],0:3])
            C[edx] = CC[best]

from itertools import product

def rotate_terminals(forest_copy):
    ANGLE_STEP = 5
    for ndx in range(forest_copy.number_of_networks):
        for edx in range(len(forest_copy.connections[ndx][0])):
            pts = []
            non_pts = []
            for tdx in range(forest_copy.trees_per_network[ndx]):
                rotated_pts = []
                non_rotated_pts = []
                T  = forest_copy.networks[ndx][tdx].data
                A  = forest_copy.assignments[ndx][tdx]
                parent = int(T[A[edx],17])
                R0 = deepcopy(T[parent,0:3])
                R1 = deepcopy(T[parent,3:6])
                P0 = deepcopy(T[A[edx],0:3])
                P1 = deepcopy(T[A[edx],3:6])
                for theta in np.arange(0,360,ANGLE_STEP):
                     non_rotated_pts.append(P0)
                     rotated_pts.append(rotate(R0,R1,P0,P1,theta))
                pts.append(np.array(rotated_pts))
                non_pts.append(np.array(non_rotated_pts))
            pts = np.array(pts)
            non_pts = np.array(non_pts)
            comb = np.array(list(product(*[range(len(rotated_pts))]*forest_copy.trees_per_network[ndx])))
            CENTERS = pts[0,comb[:,0],:]
            for tdx in range(1,forest_copy.trees_per_network[ndx]):
                CENTERS += pts[tdx,comb[:,tdx],:]
            CENTERS = CENTERS/pts.shape[0]
            ANGLES  = []
            for tdx in range(forest_copy.trees_per_network[ndx]):

                VECTOR_1 = non_pts[tdx,comb[:,tdx],:] - pts[tdx,comb[:,tdx],:]
                VECTOR_2 = CENTERS - pts[tdx,comb[:,tdx],:]
                VECTOR_1 = VECTOR_1/np.linalg.norm(VECTOR_1,axis=1).reshape(-1,1)
                VECTOR_2 = VECTOR_2/np.linalg.norm(VECTOR_2,axis=1).reshape(-1,1)
                print(VECTOR_1[0:3,:])
                print(VECTOR_2[0:3,:])
                DOT      = np.array([np.dot(VECTOR_1[i,:],VECTOR_2[i,:]) for i in range(VECTOR_1.shape[0])])
                ang      = (np.arccos(DOT)/np.pi)*180

                ANGLES.append(ang)
            ANGLES = np.array(ANGLES)
            return

def view(forest,highlight=None):
    model_networks = []
    for net_id,network in enumerate(forest.networks):
        model_trees = []
        for tree_id,network_tree in enumerate(network):
            model = []
            for edge in range(network_tree.data.shape[0]):
                if network_tree.data[edge,15]==-1 and network_tree.data[edge,16]==-1 and edge != 0:
                    dis_point = (network_tree.data[edge,0:3] + network_tree.data[edge,3:6])/2
                    vessel_length = network_tree.data[edge,20]/2
                else:
                    dis_point = network_tree.data[edge,3:6]
                    vessel_length = network_tree.data[edge,20]
                center = tuple((network_tree.data[edge,0:3] + dis_point)/2)
                radius = network_tree.data[edge,21]
                direction = tuple(network_tree.data[edge,12:15])
                model.append(pv.Cylinder(radius=radius,height=vessel_length,
                                         center=center,direction=direction,
                                         resolution=100))
            model_trees.append(model)
        model_networks.append(model_trees)
    plot = pv.Plotter()
    #colors = ['r','b','g','y']
    #plot.set_background(color=[253,250,219])
    for net_idx,model_network in enumerate(model_networks):
        for tree_idx, model_tree in enumerate(model_network):
            for vessel_idx,model in enumerate(model_tree):
                if net_idx == highlight[0] and tree_idx == highlight[1] and vessel_idx==highlight[2]:
                    plot.add_mesh(model,color='g')
                else:
                    plot.add_mesh(model,color='r')
    plot.show()
"""



import numpy as np
import pyvista as pv
import vtk
import sys,os
from .implicit.implicit import surface
from .branch_addition.check import *
from .branch_addition.close import *
from .branch_addition.basis import *
from .branch_addition.calculate_length import *
from .branch_addition.calculate_radii import *
from .branch_addition.set_root import *
from .branch_addition.add_edge import *
from .branch_addition.add_bifurcation import *
from .branch_addition.add_branches_v3 import *
from .forest_utils.connect import *
from .forest_utils.smooth import *
from .forest_utils.compete_add import *
from .sv_interface.get_sv_data import *
from .sv_interface.options import *
from .sv_interface.build_files import *
from .sv_interface.waveform import generate_physiologic_wave, wave
from .sv_interface.build_results import make_results
from .sv_interface.view_0d_result_plots import view_plots
from .sv_interface.build_0d_run_script import run_0d_script
from .sv_interface.locate import locate_0d_solver, locate_1d_solver
from .sv_interface.export_3d_only import export_3d_only
from .collision.collision import *
from .utils.gcode.gcode import *
from copy import deepcopy
from itertools import combinations
from tqdm import tqdm
import pickle
import tetgen
import json
from scipy.interpolate import splev, splprep
import matplotlib.pyplot as plt
import platform
import pymeshfix
#from wrapt_timeout_decorator import *

class tree:
    """
    Data structure:\n
    index: 0:2   -> proximal node coordinates \n
    index: 3:5   -> distal node coordinates \n
    index: 6:8   -> unit basis U (all) \n
    index: 9:11  -> unit basis V (all) \n
    index: 12:14 -> unit basis W (axial direction) (all) \n
    index: 15,16 -> children (-1 means no child) \n
    index: 17    -> parent (NA) \n
    index: 18    -> proximal node index (only real edges) \n
    index: 19    -> distal node index (only real edges) \n
    index: 20    -> length (path length) \n
    index: 21    -> radius (what are the units?) \n
    index: 22    -> flow (NA) \n
    index: 23    -> left bifurcation (NA) \n
    index: 24    -> right bifurcation (NA) \n
    index: 25    -> reduced resistance (NA) \n
    index: 26    -> depth (NA) \n
    index: 27    -> reduced downstream length (NA) \n
    index: 28    -> root radius scaling factor (same as edge for intermediate) \n
    index: 29    -> edge that subedge belongs to (if actual edge; self pointer) \n
    index: 30    -> self identifying index (-1 if intermediate) \n
    """
    def __init__(self):
        #self.parameters = {'gamma'   : 3.0,
        #                   'lambda'  : 2.0,
        #                   'mu'      : 1.0,
        #                   'nu'      : 3.6/100,
        #                   'Pperm'   : 100*1333.22,
        #                   'Pterm'   : 60*1333.22,
        #                   'Qterm'   : (0.125)/60,
        #                   'edge_num': 0}
        self.set_parameters()
        self.data = np.zeros((1,31))
        self.radius_buffer = 0.01
        self.fraction = None
        self.set_assumptions()
        self.rng_points = []
        self.time = {'search':[],
                     'constraints':[],
                     'local_optimize':[],
                     'collision':[],
                     'close_time':[],
                     'search_1':[],
                     'search_2':[],
                     'search_3':[],
                     'search_4':[],
                     'add_time':[],
                     'add_1':[],
                     'add_2':[],
                     'add_3':[],
                     'add_4':[],
                     'total':[],
                     'brute_optimize': [],
                     'method_optimize': [],
                     'depth':[],
                     'method_time':[],
                     'brute_time':[],
                     'brute_x_value':[],
                     'brute_value':[],
                     'method_x_value':[],
                     'method_value':[],
                     'truth_x_value':[],
                     'truth_value':[]}

    def set_parameters(self,**kwargs):
        self.parameters = {}
        self.parameters['gamma']    = kwargs.get('gamma',3.0)
        self.parameters['lambda']   = kwargs.get('lambda',2.0)
        self.parameters['mu']       = kwargs.get('mu',1.0)
        self.parameters['nu']       = kwargs.get('nu',3.6)/100
        self.parameters['Pperm']    = kwargs.get('Pperm',100)*1333.22
        self.parameters['Pterm']    = kwargs.get('Pterm',60)*1333.22
        self.parameters['Qterm']    = kwargs.get('Qterm',0.125)/60
        self.parameters['edge_num'] = kwargs.get('edge_num',0)
        self.parameters['rho']      = kwargs.get('rho',1.06)

    def set_assumptions(self,**kwargs):
        self.homogeneous = kwargs.get('homogeneous',True)
        self.directed    = kwargs.get('directed',False)
        self.convex      = kwargs.get('convex',False)
        self.hollow      = kwargs.get('hollow',False)
        self.dimension   = kwargs.get('dimension',3)
        self.clamped_root= kwargs.get('clamped_root',False)
        self.nonconvex_counter = 0

    def show_assumptions(self):
        print('homogeneous : {}'.format(self.homogeneous))
        print('directed    : {}'.format(self.directed))
        print('convex      : {}'.format(self.convex))
        print('hollow      : {}'.format(self.hollow))
        print('dimension   : {}'.format(self.dimension))
        print('clamped root: {}'.format(self.clamped_root))

    def set_boundary(self,boundary):
        self.boundary = boundary
        self.fraction = (self.boundary.volume**(1/3))/20

    def set_root(self,low=-1,high=0,start=None,direction=None,
                 limit_high=None,limit_low=None,niter=200):
        Qterm = self.parameters['Qterm']
        gamma = self.parameters['gamma']
        nu    = self.parameters['nu']
        Pperm = self.parameters['Pperm']
        Pterm = self.parameters['Pterm']
        result = set_root(self.data,self.boundary,Qterm,
                          gamma,nu,Pperm,Pterm,self.fraction,
                          self.homogeneous,self.convex,self.directed,
                          start,direction,limit_high,limit_low,self.boundary.volume,
                          low=-1,high=0,niter=niter)
        self.data = result[0]
        self.sub_division_map = result[1]
        self.sub_division_index = result[2]
        self.parameters['edge_num'] = 1
        #self.sub_division_map = [-1]
        #self.sub_division_index = np.array([0])

    def add(self,low,high,isforest=False,radius_buffer=0.01,threshold=None,method='L-BFGS-B'):
        vessel,data,sub_division_map,sub_division_index,threshold = add_branch(self,low,high,threshold_exponent=0.5,
                                                                     threshold_adjuster=0.75,all_max_attempts=40,
                                                                     max_attemps=3,sampling=20,max_skip=8,
                                                                     flow_ratio=None,radius_buffer=radius_buffer,
                                                                     isforest=isforest,threshold=threshold,method=method)
        if isforest:
            return vessel,data,sub_division_map,sub_division_index,threshold
        else:
            self.data = data
            self.parameters['edge_num'] += 2
            self.sub_division_map = sub_division_map
            self.sub_division_index = sub_division_index

    def n_add(self,n,method='L-BFGS-B'):
        self.rng_points,_ = self.boundary.pick(size=40*n,homogeneous=True)
        self.rng_points = self.rng_points.tolist()
        for i in tqdm(range(n),desc='Adding vessels'):
            #for i in range(n):
            #self.rng_points = self.rng_points.tolist()
            self.add(-1,0,method=method)

        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['search'][1:],label='search time')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['constraints'][1:],label='constraint time')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['local_optimize'][1:],label='local time')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['collision'][1:],label='collision time')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['search_1'][1:],label='search 1 time')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['search_2'][1:],label='search 2 time')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['search_3'][1:],label='search 3 time')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['search_4'][1:],label='search 4 time')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['close_time'][1:],label='close_time time')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['add_time'][1:],label='add_time')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['add_1'][1:],label='add_1')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['add_2'][1:],label='add_2')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['add_3'][1:],label='add_3')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['add_4'][1:],label='add_4')
        #plt.plot(list(range(len(self.time['search'][1:]))),self.time['total'][1:],label='total')
        #plt.ylim([0,0.1])
        #plt.legend()
        #plt.show()

    def show(self,surface=False,vessel_colors='red',background_color='white',
             resolution=100,show_segments=True,save=False,name=None,show=True,
             surface_color='red',other_surface_color='blue'):
        models = []
        actors = []
        if not isinstance(vessel_colors,str):
            colors = vtk.vtkColor3d()
            colors.Set(vessel_colors[0],vessel_colors[1],vessel_colors[2])
        else:
            colors = vtk.vtkNamedColors()
            colors = colors.GetColor3d(vessel_colors)
        if not isinstance(background_color,str):
            backcolors = vtk.vtkColor3d()
            backcolors.Set(background_color[0],background_color[1],background_color[2])
        else:
            backcolors = vtk.vtkNamedColors()
            backcolors = backcolors.GetColor3d(background_color)
        if not isinstance(surface_color,str):
            surf_color = vtk.vtkColor3d()
            surf_color.Set(surface_color[0],surface_color[1],surface_color[2])
        else:
            surf_color = vtk.vtkNamedColors()
            surf_color = surf_color.GetColor3d(surface_color)
        if not isinstance(other_surface_color,str):
            other_surf_color = vtk.vtkColor3d()
            other_surf_color.Set(other_surface_color[0],other_surface_color[1],other_surface_color[2])
        else:
            other_surf_color = vtk.vtkNamedColors()
            other_surf_color = other_surf_color.GetColor3d(other_surface_color)
        if show_segments:
            if self.homogeneous:
                data_subset = self.data[self.data[:,-1] > -1]
            else:
                data_subset = self.data[self.data[:,29] > -1]
            for edge in range(data_subset.shape[0]):
                center = tuple((data_subset[edge,0:3] + data_subset[edge,3:6])/2)
                radius = data_subset[edge,21]
                direction = tuple(data_subset[edge,12:15])
                vessel_length = data_subset[edge,20]
                cyl = vtk.vtkTubeFilter()
                line = vtk.vtkLineSource()
                line.SetPoint1(data_subset[edge,0],data_subset[edge,1],data_subset[edge,2])
                line.SetPoint2(data_subset[edge,3],data_subset[edge,4],data_subset[edge,5])
                cyl.SetInputConnection(line.GetOutputPort())
                cyl.SetRadius(radius)
                cyl.SetNumberOfSides(resolution)
                models.append(cyl)
                mapper = vtk.vtkPolyDataMapper()
                actor  = vtk.vtkActor()
                mapper.SetInputConnection(cyl.GetOutputPort())
                actor.SetMapper(mapper)
                actor.GetProperty().SetColor(colors)
                actors.append(actor)
            if surface:
                mapper = vtk.vtkPolyDataMapper()
                actor  = vtk.vtkActor()
                mapper.SetInputData(self.boundary.polydata)
                actor.SetMapper(mapper)
                actor.GetProperty().SetColor(surf_color)
                actor.GetProperty().SetOpacity(0.25)
                actors.append(actor)
                for other_surface in self.boundary.subtracted_volumes:
                    mapper = vtk.vtkPolyDataMapper()
                    actor  = vtk.vtkActor()
                    mapper.SetInputData(other_surface.polydata)
                    actor.SetMapper(mapper)
                    actor.GetProperty().SetColor(other_surf_color)
                    actor.GetProperty().SetOpacity(0.5)
                    actors.append(actor)
            renderer = vtk.vtkRenderer()
            renderer.SetBackground(backcolors)

            render_window = vtk.vtkRenderWindow()
            render_window.AddRenderer(renderer)
            render_window.SetWindowName('SimVascular Vessel Network')

            interactor = vtk.vtkRenderWindowInteractor()
            interactor.SetRenderWindow(render_window)

            for actor in actors:
                renderer.AddActor(actor)

            if save:
                render_window.Render()
                w2if = vtk.vtkWindowToImageFilter()
                w2if.SetInput(render_window)
                w2if.Update()

                writer = vtk.vtkPNGWriter()
                if name is None:
                    tag = time.gmtime()
                    name = str(tag.tm_year)+str(tag.tm_mon)+str(tag.tm_mday)+str(tag.tm_hour)+str(tag.tm_min)+str(tag.tm_sec)
                writer.SetFileName(os.getcwd()+os.sep+'{}.png'.format(name))
                writer.SetInputData(w2if.GetOutput())
                writer.Write()
            elif show==True:
                render_window.Render()
                interactor.Start()
            elif show==False:
                for m in models:
                    m.Update()
            return models

    def save(self,filename=None):
        if filename is None:
            tag = time.gmtime()
            filename = 'network_{}'.format(str(tag.tm_year)+
                                               str(tag.tm_mon) +
                                               str(tag.tm_mday)+
                                               str(tag.tm_hour)+
                                               str(tag.tm_min) +
                                               str(tag.tm_sec))
        os.mkdir(filename)
        file = open(filename+"/vessels.ccob",'wb')
        pickle.dump(self.data,file)
        file.close()
        file = open(filename+"/parameters.ccob",'wb')
        parameters = (self.parameters,self.fraction,self.homogeneous,self.directed,self.convex)
        pickle.dump(parameters,file)
        file.close()
        file = open(filename+"/boundary.ccob",'wb')
        pickle.dumps(self.boundary,file)
        file.close()

    def load(self,filename):
        file = open(filename+"/vessels.ccob",'rb')
        self.data = pickle.load(file)
        file.close()
        file = open(filename+"/parameters.ccob",'rb')
        self.parameters,self.fraction,self.homogeneous,self.directed,self.convex = pickle.load(file)
        file.close()
        file = open(filename+"/boundary.ccob",'rb')
        self.boundary = pickle.loads(file)
        file.close()

    def export(self,steady=True,apply_distal_resistance=True,gui=True,cylinders=False,make=True,global_edge_size=None):
        if cylinders:
            pv_data = []
            models = self.show(show=False)
            for m in models:
                pv_data.append(pv.PolyData(var_inp=m.GetOutput()))
            merge = pv_data[0].merge(pv_data[1:])
            return merge
        interp_xyz,interp_r,interp_n,frames,branches,interp_xyzr = get_interpolated_sv_data(self.data)
        points,radii,normals    = sv_data(interp_xyzr,interp_r,radius_buffer=self.radius_buffer)
        if steady:
            time = [0, 1]
            flow = [self.data[0,22], self.data[0,22]]
        else:
            time,flow = wave(self.data[0,22],self.data[0,21]*2) #changed wave function
            time = time.tolist()
            flow = flow.tolist()
            flow[-1] = flow[0]
        if apply_distal_resistance:
            R = self.parameters['Pterm']/self.data[0,22]
        else:
            R = 0
        if make:
            num_caps = 1+2+int((self.parameters['edge_num']-1)/2)
            options = file_options(num_caps,time=time,flow=flow,gui=gui,distal_resistance=R)
            if global_edge_size is None:
                options.set_mesh_options(global_edge_size=global_edge_size)
            build(points,radii,normals,options)
        return interp_xyz,interp_r,interp_xyzr

    def export_truncated(self,steady=True,radius=None,indicies=None,apply_distal_resistance=True,gui=True,cylinders=False,make=True):
        if cylinders:
            pv_data = []
            models = self.show(show=False)
            for m in models:
                pv_data.append(pv.PolyData(var_inp=m.GetOutput()))
            merge = pv_data[0].merge(pv_data[1:])
            return merge
        interp_xyz,interp_r,interp_n,frames,branches,interp_xyzr = get_truncated_interpolated_sv_data(self.data,radius=radius,indicies=indicies)
        points,radii,normals    = sv_data(interp_xyzr,interp_r,radius_buffer=self.radius_buffer)
        if steady:
            time = [0, 1]
            flow = [self.data[0,22], self.data[0,22]]
        else:
            time,flow = wave(self.data[0,22],self.data[0,21]*2)
            time = time.tolist()
            flow = flow.tolist()
            flow[-1] = flow[0]
        if apply_distal_resistance:
            R = self.parameters['Pterm']/self.data[0,22]
        else:
            R = 0
        if make:
            num_caps = 1+2+int((self.parameters['edge_num']-1)/2)
            options = file_options(num_caps,time=time,flow=flow,gui=gui,distal_resistance=R)
            build(points,radii,normals,options)
        return interp_xyz,interp_r

    def show_truncated(self,radius=None,indicies=None):
        large,small = truncate(self.data,radius=radius,indicies=indicies)
        combined = []
        plotter = pv.Plotter()
        models = []
        for i in large:
            combined += i
        for i in range(self.data.shape[0]):
            center = (self.data[i,0:3] + self.data[i,3:6])/2
            m = pv.Cylinder(center=center,direction=self.data[i,12:15],radius=self.data[i,21],height=self.data[i,20])
            if i in combined:
                plotter.add_mesh(m,color='red')
            else:
                plotter.add_mesh(m,color='blue',opacity=0.4)
        plotter.set_background('white')
        return plotter

    def export_3d_solid(self,outdir=None,folder="3d_tmp"):
        if outdir is None:
            outdir = os.getcwd()+os.sep+folder
        else:
            outdir = outdir+os.sep+folder
        if not os.path.isdir(outdir):
            os.mkdir(folder)
        def polyline_from_points(pts,r):
            poly = pv.PolyData()
            poly.points = pts
            cell = np.arange(0,len(pts),dtype=np.int_)
            cell = np.insert(cell,0,len(pts))
            poly.lines = cell
            poly.scalars = r/r[0]
            return poly
        interp_xyz,interp_r = self.export(make=False)
        vessels = []
        tets = []
        t_list = np.linspace(0,1,num=1000)
        for vessel_id in range(len(interp_xyz)):
            x,y,z = splev(t_list,interp_xyz[vessel_id][0])
            _,r = splev(t_list,interp_r[vessel_id][0])
            points = np.zeros((len(t_list),3))
            points[:,0] = x
            points[:,1] = y
            points[:,2] = z
            polyline = polyline_from_points(points,r)
            vessel = polyline.tube(radius=r[0])
            vessels.append(vessel)
            tets.append(tetgen.TetGen(vessel.triangulate()))
        for i in tqdm(range(len(tets)),desc="Tetrahedralizing"):
            tets[i].tetrahedralize(minratio=1.2)
        surfs = []
        for i in tqdm(range(len(tets)),desc="Extracting Surfaces"):
            surfs.append(tets[i].grid.extract_surface().triangulate())
        unioned = surfs[0].boolean_union(surfs[1])
        unioned = unioned.clean()
        for i in tqdm(range(2,len(surfs)),desc="Unioning"):
            unioned = unioned.boolean_union(surfs[i])
            unioned = unioned.clean()
        unioned.save(outdir+os.sep+"unioned_solid.vtp")
        with open(outdir+os.sep+"simvascular_python_script.py","w") as file:
            file.write(export_3d_only.format(outdir+os.sep+"unioned_solid.vtp"))
        return vessels,tets,unioned

    def export_0d_simulation(self,steady=True,outdir=None,folder="0d_tmp",number_cardiac_cycles=1,
                      number_time_pts_per_cycle=5,density=1.06,viscosity=0.04,material="olufsen",
                      olufsen={'k1':0.0,'k2':-22.5267,'k3':1.0e7,'material exponent':1.0,'material pressure':0.0},
                      linear={'material ehr':1e7,'material pressure':0.0},path_to_0d_solver=None,
                      viscosity_model='constant',vivo=True):
        """
        This script builds the 0D input file for running 0D simulation.
        Parameters:
        -----------
        steady: bool, optional, [default=True]
        outdir: str, optional, [default=None]
        folder: str, optional, [default="tmp"]
        number_cardiac_cycles: int, optional, [default=1]
        """
        if outdir is None:
            outdir = os.getcwd()+os.sep+folder
        else:
            outdir = outdir+os.sep+folder
        if not os.path.isdir(outdir):
            os.mkdir(folder)
        if path_to_0d_solver is None:
            path_to_0d_solver = locate_0d_solver()
        else:
            path_to_0d_solver = locate_0d_solver(windows_drive=path_to_0d_solver,linux_drive=path_to_0d_solver)
        input_file = {'description':{'description of case':None,
                                     'analytical results':None},
                      'boundary_conditions':[],
                      'junctions':[],
                      'simulation_parameters':{},
                      'vessels':[]}
        simulation_parameters = {}
        simulation_parameters["number_of_cardiac_cycles"] = number_cardiac_cycles
        simulation_parameters["number_of_time_pts_per_cardiac_cycle"] = number_time_pts_per_cycle
        input_file['simulation_parameters'] = simulation_parameters
        for vessel in range(self.data.shape[0]):
            tmp = {}
            tmp['vessel_id'] = vessel
            tmp['vessel_length'] = self.data[vessel,20]
            tmp['vessel_name'] = 'branch'+str(vessel)+"_seg0"
            tmp['zero_d_element_type'] = "BloodVessel"
            if material == "olufsen":
                material_stiffness = olufsen['k1']*np.exp(olufsen['k2']*self.data[vessel,21])+olufsen['k3']
            else:
                material_stiffness = linear['material ehr']
            zero_d_element_values = {}
            if viscosity_model == 'constant':
                nu = self.parameters['nu']
            elif viscosity_model == 'modified viscosity law':
                W  = 1.1
                lam = 0.5
                D  = self.data[vessel,21]*2*(10000)
                nu_ref = self.parameters['nu']
                Hd = 0.45 #discharge hematocrit (dimension-less)
                C  = lambda d: (0.8+np.exp(-0.075))*(-1+(1/(1+10**(-11)*d**12)))+(1/(1+10**(-11)*d**12))
                nu_vitro_45 = lambda d: 220*np.exp(-1.3*d) + 3.2-2.44*np.exp(-0.06*d**0.645)
                nu_vivo_45  = lambda d: 6*np.exp(-0.085*d)+3.2-2.44*np.exp(-0.06*d**0.645)
                if vivo == True:
                    nu_45 = nu_vivo_45
                else:
                    nu_45 = nu_vitro_45
                nu_mod   =   lambda d: (1+(nu_45 - 1)*(((1-Hd)**C-1)/((1-0.45)**C-1))*(d/(d-W))**(4*lam))*(d/(d-W))**(4*(1-lam))
                ref = nu_ref - nu_mod(10000) # 1cm (units given in microns)
                nu  = nu_mod(D) + ref
            zero_d_element_values["R_poiseuille"] = ((8*nu/np.pi)*self.data[vessel,20])/self.data[vessel,21]**4
            zero_d_element_values["C"] = (3*self.data[vessel,20]*np.pi*self.data[vessel,21]**2)/(2*material_stiffness)
            zero_d_element_values["L"] = (self.data[vessel,20]*density)/(np.pi*self.data[vessel,21]**2)
            zero_d_element_values["stenosis_coefficient"] = 0.0
            tmp['zero_d_element_values'] = zero_d_element_values
            if vessel == 0:
                bc = {}
                bc['bc_name'] = "INFLOW"
                bc['bc_type'] = "FLOW"
                bc_values = {}
                if steady:
                    bc_values["Q"] = [self.data[vessel,22], self.data[vessel,22]]
                    bc_values["t"] = [0, 1]
                    with open(outdir+os.sep+"inflow.flow","w") as file:
                        for i in range(len(bc_values["t"])):
                            file.write("{}  {}\n".format(bc_values["t"][i],bc_values["Q"][i]))
                    file.close()
                else:
                    time,flow = wave(self.data[vessel,22],self.data[vessel,21]*2) # changed wave function
                    bc_values["Q"] = flow.tolist()
                    bc_values["t"] = time.tolist()
                    bc_values["Q"][-1] = bc_values["Q"][0]
                    simulation_parameters["number_of_time_pts_per_cardiac_cycle"] = len(bc_values["Q"])
                    with open(outdir+os.sep+"inflow.flow","w") as file:
                        for i in range(len(bc_values["t"])):
                            file.write("{}  {}\n".format(bc_values["t"][i],bc_values["Q"][i]))
                    file.close()
                bc['bc_values'] = bc_values
                input_file['boundary_conditions'].append(bc)
                tmp['boundary_conditions'] = {'inlet':"INFLOW"}
                if self.data[vessel,15] > 0 and self.data[vessel,16] > 0:
                    junction = {}
                    junction['inlet_vessels'] = [vessel]
                    junction['junction_name'] = "J"+str(vessel)
                    junction['junction_type'] = "NORMAL_JUNCTION"
                    junction['outlet_vessels'] = [int(self.data[vessel,15]), int(self.data[vessel,16])]
                    input_file['junctions'].append(junction)
            elif self.data[vessel,15] < 0 and self.data[vessel,16] < 0:
                bc = {}
                bc['bc_name'] = "OUT"+str(vessel)
                bc['bc_type'] = "RESISTANCE"
                bc_values = {}
                bc_values["Pd"] = self.parameters["Pterm"]
                bc_values["R"] = 0
                bc['bc_values'] = bc_values
                input_file['boundary_conditions'].append(bc)
                tmp['boundary_conditions'] = {'outlet':'OUT'+str(vessel)}
            else:
                junction = {}
                junction['inlet_vessels'] = [vessel]
                junction['junction_name'] = "J"+str(vessel)
                junction['junction_type'] = "NORMAL_JUNCTION"
                junction['outlet_vessels'] = [int(self.data[vessel,15]), int(self.data[vessel,16])]
                input_file['junctions'].append(junction)
            input_file['vessels'].append(tmp)
        obj = json.dumps(input_file,indent=4)
        with open(outdir+os.sep+"solver_0d.in","w") as file:
            file.write(obj)
        file.close()

        with open(outdir+os.sep+"plot_0d_results_to_3d.py","w") as file:
            file.write(make_results)
        file.close()

        with open(outdir+os.sep+"plot_0d_results_at_slices.py","w") as file:
            file.write(view_plots)
        file.close()

        with open(outdir+os.sep+"run.py","w") as file:
            if platform.system() == "Windows":
                if path_to_0d_solver is not None:
                    solver_path = path_to_0d_solver.replace(os.sep,os.sep+os.sep)
                else:
                    solver_path = path_to_0d_solver
                    print("WARNING: Solver location will have to be given manually")
                    print("Current solver path is: {}".format(solver_path))
                solver_file = (outdir+os.sep+"solver_0d.in").replace(os.sep,os.sep+os.sep)
            else:
                if path_to_0d_solver is not None:
                    solver_path = path_to_0d_solver
                else:
                    solver_path = path_to_0d_solver
                    print("WARNING: Solver location will have to be given manually")
                    print("Current solver path is: {}".format(solver_path))
                solver_file = outdir+os.sep+"solver_0d.in"
            file.write(run_0d_script.format(solver_path,solver_file))
        file.close()

        geom = np.zeros((self.data.shape[0],8))
        geom[:,0:3] = self.data[:,0:3]
        geom[:,3:6] = self.data[:,3:6]
        geom[:,6]   = self.data[:,20]
        geom[:,7]   = self.data[:,21]
        np.savetxt(outdir+os.sep+"geom.csv",geom,delimiter=",")
    def collision_free(self,outside_vessels,radius_buffer=0.01):
        return no_outside_collision(self,outside_vessels,radius_buffer=radius_buffer)

    def export_gcode(self):
        interp_xyz,interp_r,interp_n,frames,branches,interp_xyzr = get_interpolated_sv_data(self.data)
        points,radii,normals    = sv_data(interp_xyzr,interp_r,radius_buffer=self.radius_buffer)
        generate_gcode(points)

class forest:
    def __init__(self,boundary=None,number_of_networks=1,
                 trees_per_network=[2],scale=None,start_points=None,
                 directions=None,root_lengths_high=None,
                 root_lengths_low=None,convex=False,compete=False):
        self.networks    = []
        self.connections = []
        self.assignments = []
        self.backup      = []
        self.boundary    = boundary
        self.convex      = convex
        self.compete     = compete
        self.number_of_networks = number_of_networks
        self.trees_per_network = trees_per_network
        if isinstance(start_points,type(None)):
            start_points = [[None for j in range(trees_per_network[i])] for i in range(number_of_networks)]
        if isinstance(directions,type(None)):
            directions = [[None for j in range(trees_per_network[i])] for i in range(number_of_networks)]
        if isinstance(root_lengths_high,type(None)):
            root_lengths_high = [[None for j in range(trees_per_network[i])] for i in range(number_of_networks)]
        if isinstance(root_lengths_low,type(None)):
            root_lengths_low = [[None for j in range(trees_per_network[i])] for i in range(number_of_networks)]
        self.starting_points = start_points
        self.directions = directions
        self.root_lengths_high = root_lengths_high
        self.root_lengths_low  = root_lengths_low
        networks = []
        for idx in range(self.number_of_networks):
            network = []
            for jdx in range(self.trees_per_network[idx]):
                tmp = tree()
                tmp.set_assumptions(convex=self.convex)
                tmp.set_boundary(self.boundary)
                if self.directions[idx][jdx] is not None:
                    tmp.clamped_root = True
                network.append(tmp)
            networks.append(network)
        self.networks    = networks

    def set_roots(self,scale=None,bounds=None):
        if self.compete:
            volume = deepcopy(self.boundary.volume)
            self.boundary.volume = self.boundary.volume/self.number_of_networks
        if self.boundary is None:
            print("Need to assign boundary!")
            return
        networks = []
        connections = []
        backup = []
        for idx in range(self.number_of_networks):
            network = []
            if not isinstance(self.trees_per_network,list):
                print("trees_per_network must be list of ints"+
                      " with length equal to number_of_networks")
            for jdx in range(self.trees_per_network[idx]):
                #tmp = tree()
                tmp = self.networks[idx][jdx]
                #tmp.set_assumptions(convex=self.convex)
                #if self.directions[idx][jdx] is not None:
                #    tmp.clamped_root = True
                #tmp.set_boundary(self.boundary)
                if scale is not None:
                    tmp.parameters['Qterm'] *= scale
                if idx == 0 and jdx == 0:
                    collisions = [True]
                else:
                    collisions = [True]
                while any(collisions):
                    if bounds is None:
                        tmp.set_root(start=self.starting_points[idx][jdx],
                                     direction=self.directions[idx][jdx],
                                     limit_high=self.root_lengths_high[idx][jdx],
                                     limit_low=self.root_lengths_low[idx][jdx])
                    else:
                        print("Not implemented yet!")
                    collisions = []
                    repeat = False
                    for net in networks:
                        for t in net:
                            collisions.append(not t.collision_free(tmp.data[0,:].reshape(1,-1)))
                            if collisions[-1]:
                                repeat = True
                                break
                        if repeat:
                            break
                    for t in network:
                        collisions.append(not t.collision_free(tmp.data[0,:].reshape(1,-1)))
                        if collisions[-1]:
                            repeat = True
                            break
                    #print(collisions)
                network.append(tmp)
            networks.append(network)
            connections.append(None)
            backup.append(None)
        self.networks    = networks
        self.connections = connections
        self.backup      = backup
        if self.compete:
            self.boundary.volume = volume

    def add(self,number_of_branches,network_id=0,radius_buffer=0.01,exact=True):
        if self.compete:
            network_id = -1
        if network_id == -1:
            exit_number = []
            active_networks = list(range(len(self.networks)))
            for network in self.networks:
                exit_number.append(network[0].parameters['edge_num'] + number_of_branches)
        else:
            exit_number = []
            active_networks = [network_id]
            for nid in range(self.number_of_networks):
                if nid in active_networks:
                    exit_number.append(self.networks[nid][0].parameters['edge_num'] + number_of_branches)
                else:
                    exit_number.append(self.networks[nid][0].parameters['edge_num'])
        for AN in active_networks:
            for ATr in range(self.trees_per_network[AN]):
                self.networks[AN][ATr].rng_points,_ = self.boundary.pick(size=40*number_of_branches,homogeneous=True)
                self.networks[AN][ATr].rng_points = self.networks[AN][ATr].rng_points.tolist()
                if self.compete:
                    # new
                    #self.networks[AN][ATr].rng_points,_ = self.boundary.pick(size=len(self.boundary.tet_verts),homogeneous=True,replacement=False)
                    #self.networks[AN][ATr].rng_points = self.networks[AN][ATr].rng_points.tolist()
                    for n in tqdm(range(len(self.networks[AN][ATr].rng_points))):
                        pt = np.array(self.networks[AN][ATr].rng_points.pop(0))
                        if exact:
                            other_vessels,distances = close_exact(self.networks[AN][ATr].data,pt)
                        else:
                            other_vessels,distances = close(self.networks[AN][ATr].data,pt)
                        minimum_distance = min(distances)
                        #if minimum_distance < 4*forest.networks[nid][njd].data[other_vessels[0],21]:
                        #    continue
                        retry = False
                        for idx in active_networks:
                            for jdx in list(range(self.trees_per_network[idx])):
                                if idx == AN:
                                    continue
                                if exact:
                                    other_vessels,distances = close_exact(self.networks[idx][jdx].data,pt)
                                else:
                                    other_vessels,distances = close(self.networks[idx][jdx].data,pt)
                                if min(distances) < minimum_distance:
                                    retry = True
                                    break
                            if retry:
                                break
                        if retry:
                            continue
                        else:
                            self.networks[AN][ATr].rng_points.insert(-1,pt.tolist())
                    #new
        if not self.compete:
            while len(active_networks) > 0:
                for nid in active_networks:
                    number_of_trees = len(self.networks[nid])
                    for njd in range(number_of_trees):
                        success = False
                        threshold = None
                        while not success:
                            vessel,data,sub_division_map,sub_division_index,threshold = self.networks[nid][njd].add(-1,0,isforest=True,radius_buffer=radius_buffer,threshold=threshold)
                            new_vessels = np.vstack((data[vessel,:],data[-2,:],data[-1,:]))
                            repeat = False
                            for nikd in range(len(self.networks)):
                                if nikd == nid:
                                    check_trees = list(range(len(self.networks[nikd])))
                                    check_trees.remove(njd)
                                else:
                                    check_trees = list(range(len(self.networks[nikd])))
                                for njkd in check_trees:
                                    if not self.networks[nikd][njkd].collision_free(new_vessels,radius_buffer=radius_buffer):
                                        repeat = True
                                        break
                                if repeat:
                                    break
                            if repeat:
                                #p = self.show()
                                #p.show()
                                #print("\n\ncollision\n\n")
                                continue
                            else:
                                success = True
                        self.networks[nid][njd].data = data
                        self.networks[nid][njd].parameters['edge_num'] += 2
                        self.networks[nid][njd].sub_division_map = sub_division_map
                        self.networks[nid][njd].sub_division_index = sub_division_index
                    #print(self.networks[nid][0].parameters['edge_num'])
                    if self.networks[nid][0].parameters['edge_num'] >= exit_number[nid]:
                        active_networks.remove(nid)
        else:
            for i in tqdm(range(number_of_branches),desc="Adding branches"):
                compete_add(self,network_ids=network_id,radius_buffer=radius_buffer)

    def show(self,show=True,resolution=100,final=False,merged_trees=False):
        if merged_trees:
            colors = ['r','b','g','y']
            plotter = pv.Plotter()
            merged_list = []
            for i in range(self.number_of_networks):
                for j in range(self.trees_per_network[i]):
                    merged_list.append(self.networks[i][j].export(cylinders=True))
                    plotter.add_mesh(merged_list[-1],colors[i])
            return plotter,merged_list
        model_networks = []
        for net_id,network in enumerate(self.networks):
            model_trees = []
            for tree_id,network_tree in enumerate(network):
                model = []
                for edge in range(network_tree.parameters['edge_num']):
                    if network_tree.data[edge,15]==-1 and network_tree.data[edge,16]==-1 and edge != 0:
                        dis_point = (network_tree.data[edge,0:3] + network_tree.data[edge,3:6])/2
                        vessel_length = network_tree.data[edge,20]/2
                    else:
                        dis_point = network_tree.data[edge,3:6]
                        vessel_length = network_tree.data[edge,20]
                    center = tuple((network_tree.data[edge,0:3] + dis_point)/2)
                    radius = network_tree.data[edge,21]
                    direction = tuple(network_tree.data[edge,12:15])
                    #vessel_length = network_tree.data[edge,20]
                    model.append(pv.Cylinder(radius=radius,height=vessel_length,
                                             center=center,direction=direction,
                                             resolution=resolution))
                if not np.any(self.connections[net_id] is None):
                    for edge in range(self.connections[net_id][tree_id].shape[0]):
                        term = self.connections[net_id][tree_id][edge,:]
                        center = tuple((term[0:3]+term[3:6])/2)
                        radius = term[21]
                        direction = term[12:15]
                        vessel_length = term[20]
                        model.append(pv.Cylinder(radius=radius,height=vessel_length,
                                                 center=center,direction=direction,
                                                 resolution=resolution))
                    #for edge in range(len(self.connections[net_id][0])):
                    #    term = network_tree.data[self.assignments[net_id][tree_id][edge],:]
                    #    conn = self.connections[net_id][0][edge]
                    #    center = tuple((term[3:6] + conn)/2)
                    #    radius = term[21]
                    #    direction = tuple(conn-term[3:6])
                    #    vessel_length = np.linalg.norm(conn-term[3:6])
                    #    model.append(pv.Cylinder(radius=radius,height=vessel_length,
                    #                             center=center,direction=direction,
                    #                             resolution=resolution))
                model_trees.append(model)
            model_networks.append(model_trees)
        """
        model_connections = []
        for connections in conn:
            if connections is None:
                continue
            else:
                for c_idx in range(connections.shape[0]):
                    center = tuple((connections[c_idx,0:3] + connections[c_idx,3:6])/2)
                    radius = connections[c_idx,21]
                    direction = tuple((connections[c_idx,3:6] - connections[c_idx,0:3])/connections[c_idx,20])
                    vessel_length = connections[c_idx,20]
                    model_conn.append(pv.Cylinder(radius=radius,height=vessel_length,
                                                  center=center,direction=direction,
                                                  resolution=resolution))
        """
        if show:
            plot = pv.Plotter()
            colors = ['r','b','g','y']
            #plot.set_background(color=[253,250,219])
            for model_idx,model_network in enumerate(model_networks):
                for color_idx, model_tree in enumerate(model_network):
                    for model in model_tree:
                        plot.add_mesh(model,colors[color_idx])
            #for c_model in model_conn:
            #    plot.add_mesh(c_model,'r')
                #if i == 0:
                #    plot.add_mesh(model[i],'r')
                #else:
                #    plot.add_mesh(model[i],'b')
            #path = plot.generate_orbital_path(n_points=100,viewup=(1,1,1))
            #plot.show(auto_close=False)
            #plot.open_gif('cco_gamma.gif')
            #plot.orbit_on_path(path,write_frames=True)
            #plot.close()
            return plot

    def connect(self,network_id=-1,buffer=None,curve_sample_size_min=5,curve_sample_size_max=11,curve_degree=3):
        self.forest_copy = self.copy()
        self.forest_copy.connections,self.forest_copy.assignments = connect(self.forest_copy,network_id=network_id,buffer=buffer)
        self.forest_copy.connections,self.forest_copy.connected_forest,self.splines = smooth(self.forest_copy,curve_sample_size_min=curve_sample_size_min,curve_sample_size_max=curve_sample_size_max,curve_degree=curve_degree)

    def assign(self):
        self.connections,self.assignments = connect(self)

    def rotate(self):
        forest_copy = self.copy()
        forest_copy.assign()
        comb,pts = rotate_terminals(forest_copy)
        self.forest_copy = forest_copy
        return comb,pts

    def copy(self):
        forest_copy = forest(boundary=self.boundary,number_of_networks=self.number_of_networks,
                             trees_per_network=self.trees_per_network,scale=None,convex=self.convex,
                             compete=self.compete)
        for ndx in range(self.number_of_networks):
            for tdx in range(self.trees_per_network[ndx]):
                forest_copy.networks[ndx][tdx].data = deepcopy(self.networks[ndx][tdx].data)
                forest_copy.networks[ndx][tdx].parameters = deepcopy(self.networks[ndx][tdx].parameters)
                forest_copy.networks[ndx][tdx].sub_division_map = deepcopy(self.networks[ndx][tdx].sub_division_map)
                forest_copy.networks[ndx][tdx].sub_division_index = deepcopy(self.networks[ndx][tdx].sub_division_index)
        forest_copy.connections = deepcopy(self.connections)
        forest_copy.assignments = deepcopy(self.assignments)
        return forest_copy

    def export_solid(self,outdir=None,folder="3d_tmp"):
        if outdir is None:
            outdir = os.getcwd()+os.sep+folder
        else:
            outdir = outdir+os.sep+folder
        if not os.path.isdir(outdir):
            os.mkdir(folder)
        def polyline_from_points(pts,r):
            poly = pv.PolyData()
            poly.points = pts
            cell = np.arange(0,len(pts),dtype=np.int_)
            cell = np.insert(cell,0,len(pts))
            poly.lines = cell
            poly["radius"] = r #r/min(r)
            return poly
        final_points,final_radii,final_normals,CONNECTED_COPY,ALL_INTERP_XYZ,ALL_INTERP_RADII = self.export()
        ALL_vessels = []
        ALL_tets = []
        ALL_surfs = []
        t_list = np.linspace(0,1,num=1000)
        P = self.show()
        for net_id in range(len(final_points)):
            for tree in range(self.trees_per_network[net_id]-1):
                interp_xyz = ALL_INTERP_XYZ[net_id][tree]
                interp_r = ALL_INTERP_RADII[net_id][tree]
                vessels = []
                tets = []
                surfs = []
                for vessel_id in range(len(self.ALL_POINTS[net_id][tree])):
                    #x,y,z = splev(t_list,interp_xyz[vessel_id][0])
                    #_,r = splev(t_list,interp_r[vessel_id][0])
                    #points = np.zeros((len(t_list),3))
                    #points[:,0] = x
                    #points[:,1] = y
                    #points[:,2] = z
                    points = np.array(self.ALL_POINTS[net_id][tree][vessel_id])
                    r = np.array(self.ALL_RADII[net_id][tree][vessel_id])
                    #print("Points: {}".format(points))
                    #print("Radii:  {}".format(r))
                    polyline = polyline_from_points(points,r)
                    vessel = polyline.tube(radius=min(r),scalars='radius',radius_factor=max(r)/min(r)).triangulate()
                    vessels.append(vessel)
                    P.add_mesh(vessel,show_edges=True,opacity=0.4)
                    P.add_points(points,render_points_as_spheres=True,point_size=20)
                    tets.append(tetgen.TetGen(vessel))
                desc1 = 'Tetrahedralizing'
                desc1 = desc1+' '*(40-len(desc1))

                # Disable
                def blockPrint():
                    sys.stdout = open(os.devnull, 'w')

                # Restore
                def enablePrint():
                    sys.stdout = sys.__stdout__

                #@timeout(60)
                def tetrahedralize(tet):
                    tet.tetrahedralize(minratio=1.2)
                    return tet

                tet_success = True
                tet_fail_list = []
                for i in tqdm(range(len(tets)),position=0,leave=False,bar_format='{l_bar}{bar:50}{r_bar}{bar:-10b}',desc=desc1):
                    blockPrint()
                    try:
                        tets[i] = tetrahedralize(tets[i])
                    except:
                        try:
                            fixed_mesh = pymeshfix.MeshFix(vessels[i])
                            fixed_mesh.repair(verbose=False)
                            tets[i] = tetgen.TetGen(fixed_mesh.mesh)
                            tets[i] = tetrahedralize(tets[i])
                        except:
                            tet_success = False
                            tet_fail_list.append(i)
                    enablePrint()
                desc2 = 'Extracting Surfaces'
                desc2 = desc2+' '*(40-len(desc2))
                surf_success = True
                surf_fail_list = []
                for i in tqdm(range(len(tets)),position=0,leave=False,bar_format='{l_bar}{bar:50}{r_bar}{bar:-10b}',desc=desc2):
                    try:
                        surfs.append(tets[i].grid.extract_surface().triangulate())
                    except:
                        surf_success = False
                        surf_fail_list.append(i)
                ALL_vessels.append(vessels)
                ALL_tets.append(tets)
                ALL_surfs.append(surfs)
                unioned = surfs[0].boolean_union(surfs[1])
                unioned = unioned.clean()
                desc3 = 'Unioning'
                desc3 = desc3+' '*(40-len(desc3))
                union_success = True
                union_fail_list = []
                last_unioned = deepcopy(unioned)
                for i in tqdm(range(2,len(surfs)),position=0,leave=False,bar_format='{l_bar}{bar:50}{r_bar}{bar:-10b}',desc=desc3):
                    try:
                        unioned = unioned.boolean_union(surfs[i])
                        unioned = unioned.clean()
                        last_unioned = deepcopy(unioned)
                    except:
                        union_success = False
                        union_fail_list.append(i)
                    unioned = last_unioned

                if union_success:
                    unioned.save(outdir+os.sep+"unioned_solid_network_{}_tree_{}.vtp".format(net_id,tree))
                else:
                    unioned.save(outdir+os.sep+"partial_unioned_solid_network_{}_tree_{}.vtp".format(net_id,tree))
                if tet_success:
                    print('Tetrahedralize    : PASS')
                else:
                    print('Tetrahedralize    : Fail ---> {}'.format(tet_fail_list))
                if surf_success:
                    print('Surface Extraction: PASS')
                else:
                    print('Surface Extraction: Fail ---> {}'.format(surf_fail_list))
                if tet_success:
                    print('Union             : PASS')
                else:
                    print('Union             : Fail ---> {}'.format(union_fail_list))
        return ALL_vessels,ALL_tets,ALL_surfs,P

    def export(self,steady=True,apply_distal_resistance=False,gui=False):
        ALL_POINTS  = []
        ALL_RADII   = []
        ALL_NORMALS = []
        ALL_INTERP_XYZ = []
        ALL_INTERP_RADII = []
        CONNECTED_COPY = []
        desc1 = 'Extracting Data from Networks'
        desc1 = desc1+' '*(40-len(desc1))
        for network in tqdm(range(self.number_of_networks),bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',position=0,leave=False,desc=desc1):
            network_points  = []
            network_radii   = []
            network_normals = []
            network_endpoints = []
            network_copy = []
            for tr in range(self.trees_per_network[network]):
                tree_copy = deepcopy(self.networks[network][tr].data)
                tree_copy_terminals = tree_copy[np.argwhere(tree_copy[:,15]==-1).flatten(),:]
                tree_copy_terminals = tree_copy_terminals[np.argwhere(tree_copy_terminals[:,16]==-1).flatten(),-1].astype(int)
                tree_copy[tree_copy_terminals,3:6] = (tree_copy[tree_copy_terminals,0:3] + tree_copy[tree_copy_terminals,3:6])/2
                tree_copy[tree_copy_terminals,20] = tree_copy[tree_copy_terminals,20]/2
                max_vessel_id = max(tree_copy[:,-1])
                conn_parents = np.argwhere(self.connections[network][tr][:,17] <= max_vessel_id).flatten()
                tree_copy[self.connections[network][tr][conn_parents,17].astype(int),15] = self.connections[network][tr][conn_parents,-1]
                tree_copy = np.vstack((tree_copy,self.connections[network][tr]))
                network_copy.append(tree_copy)
                #tmp_tree = tree()
                #tmp_tree.data = tree_copy
                #tmp_tree.show()

            for tr,net in enumerate(network_copy):
                update(net,self.networks[network][tr].parameters['gamma'],self.networks[network][tr].parameters['nu'])
                update_radii(net,self.networks[network][tr].parameters['Pperm'],self.networks[network][tr].parameters['Pterm'])
                #tmp_tree = tree()
                #tmp_tree.data = net
                #tmp_tree.show()
            #P = self.show()
            total = self.trees_per_network[network]
            desc2 = 'Extracting Data from Network {} Trees'.format(network)
            desc2 = desc2+' '*(40-len(desc2))
            with tqdm(total=total,bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}',position=1,leave=False,desc=desc2) as pbar_tree:
                for tree_copy in network_copy:
                    interp_xyz,interp_r,interp_n,frames,branches,interp_xyzr = get_interpolated_sv_data(tree_copy)
                    points,radii,normals = sv_data(interp_xyzr,interp_r,radius_buffer=0.01)
                    #for p_list in points:
                    #    P.add_points(np.array(p_list),render_points_as_spheres=True,color='g',point_size=30)
                    network_points.append(points)
                    network_radii.append(radii)
                    network_normals.append(normals)
                    pts = [p[-1] for p in points]
                    network_endpoints.append(pts)
                    #P.show()
                    pbar_tree.update(1)
            CONNECTED_COPY.append(network_copy)

            # Calculate Vessel Reordering relative to the first vessel of the network
            order = [list(range(len(network_endpoints[0])))]
            for tree in range(1,self.trees_per_network[network]):
                tmp = []
                for idx in order[0]:
                    #tmp.append(np.argwhere(np.all(np.isclose(e[0][0],network_endpoints[]),axis=1)).flatten()[0])
                    tmp.append(network_endpoints[tree].index(network_endpoints[0][idx]))
                order.append(tmp)
            # Reorder Vessels
            reordered_points  = []
            reordered_radii   = []
            reordered_normals = []
            for tree in range(self.trees_per_network[network]):
                tmp_points  = []
                tmp_radii   = []
                tmp_normals = []
                for idx in order[tree]:
                    tmp_points.append(network_points[tree][idx])
                    tmp_radii.append(network_radii[tree][idx])
                    tmp_normals.append(network_normals[tree][idx])
                reordered_points.append(tmp_points)
                reordered_radii.append(tmp_radii)
                reordered_normals.append(tmp_normals)
                #print('reorder: {}'.format(len(reordered_points[-1])))
            # Reverse and Combine
            final_points  = [reordered_points[0]]
            final_radii   = [reordered_radii[0]]
            final_normals = [reordered_normals[0]]
            final_interp_xyz = []
            final_interp_radii = []
            for tree in range(1,self.trees_per_network[network]):
                if tree == 1:
                    for vessel in range(len(reordered_points[tree])):
                        vessel_normals_flip = np.array(reordered_normals[tree][vessel])*-1
                        reordered_normals[tree][vessel] = vessel_normals_flip.tolist()
                        final_points[0][vessel].extend(list(reversed(reordered_points[tree][vessel]))[2:]) #changed
                        final_radii[0][vessel].extend(list(reversed(reordered_radii[tree][vessel]))[2:]) #changed
                        final_normals[0][vessel].extend(list(reversed(reordered_normals[tree][vessel]))[2:]) #changed 9-8-22
                        #make the connecting contours have equal average radii
                else:
                    tmp_points  = []
                    tmp_radii   = []
                    tmp_normals = []
                    for vessel in range(len(reordered_points[tree])):
                        tmp_points.append(list(reversed(reordered_points[tree][vessel])))
                        tmp_radii.append(list(reversed(reordered_radii[tree][vessel])))
                        tmp_normals.append(list(reversed(reordered_normals[tree][vessel])))
                        # make sure the radii is less than the average radii of the first vessel
                    final_points.append(tmp_points)
                    final_radii.append(tmp_radii)
                    final_normals.append(tmp_normals)
                interp_xyz  = []
                interp_r    = []
                for vessel in range(len(final_points[-1])):
                    pass
                    #P.add_points(np.array(final_points[-1][vessel]),render_points_as_spheres=True,color='g',point_size=30)
                    #interp_xyz.append(splprep(np.array(final_points[-1][vessel]).T,s=0))
                    #r = np.vstack((interp_xyz[-1][1],np.array(final_radii[-1][vessel]).T))
                    #interp_r.append(splprep(r,s=0))
                final_interp_xyz.append(interp_xyz)
                final_interp_radii.append(interp_r)
            #P.show()
            ALL_POINTS.append(final_points)
            ALL_RADII.append(final_radii)
            ALL_NORMALS.append(final_normals)
            ALL_INTERP_XYZ.append(final_interp_xyz)
            ALL_INTERP_RADII.append(final_interp_radii)
            self.ALL_POINTS = ALL_POINTS
            self.ALL_RADII = ALL_RADII
            # Export Components of Network to SV Builder for Code Generation
            for component in range(len(final_points)):
                if steady:
                    time = [0, 1]
                    flow = [self.networks[network][0].data[0,22], self.networks[network][0].data[0,22]]
                else:
                    time,flow = wave(self.networks[network][0].data[0,22],self.networks[network][0].data[0,21]*2) # changed wave function
                    time = time.tolist()
                    flow = flow.tolist()
                    flow[-1] = flow[0]
                if apply_distal_resistance:
                    R = self.networks[network][0].parameters['Pterm']/self.networks[network][0].data[0,22]
                else:
                    R = 0
                num_caps = self.trees_per_network[network]
                options = file_options(num_caps,time=time,flow=flow,gui=gui,distal_resistance=R)
                build(final_points[component],final_radii[component],final_normals[component],options)

        return final_points,final_radii,final_normals,CONNECTED_COPY,ALL_INTERP_XYZ,ALL_INTERP_RADII

def perfusion_territory(tree,subdivisions=0,mesh_file=None,tree_file=None):
    terminals = tree.data[tree.data[:,15]<0,:]
    terminals = terminals[terminals[:,16]<0,:]
    territory_id = []
    territory_volumes = np.zeros(len(terminals))
    vol = tree.boundary.tet.grid.compute_cell_sizes().cell_data['Volume']
    #for idx, cell_center in tqdm(enumerate(tree.boundary.tet.grid.cell_centers().points),desc='Calculating Perfusion Territories'):
    for idx, cell_center in enumerate(tree.boundary.tet.grid.cell_centers().points):
        closest_terminal = np.argmin(np.linalg.norm(cell_center.reshape(1,-1)-terminals[:,3:6],axis=1))
        territory_id.append(closest_terminal)
        territory_volumes[closest_terminal] += vol[idx]
    territory_id = np.array(territory_id)
    tree.boundary.tet.grid['perfusion_territory_id'] = territory_id
    if mesh_file is not None:
        tree.boundary.tet.grid.save(mesh_file)
    else:
        tree.boundary.tet.grid.save('example.vtu')
    models = tree.show(show=False)
    append = vtk.vtkAppendPolyData()
    #append.UserManagedInputsOn()
    #append.SetInputDataObject(models[0])
    #append.SetNumberOfInputs(len(models))
    for i in range(len(models)):
        append.AddInputData(models[i].GetOutput())
    append.Update()
    total_tree = append.GetOutput()
    writer = vtk.vtkXMLPolyDataWriter()
    if tree_file is not None:
        writer.SetFileName(tree_file)
    else:
        writer.SetFileName(os.getcwd()+os.sep+'example_tree.vtp')
    writer.SetInputDataObject(total_tree)
    writer.Update()
    return territory_id,territory_volumes/np.sum(territory_volumes)



import tkinter as tk
from tkinter import ttk

class OptionsEntry:
    def __init__(self):
        main = tk.Tk()
        main.title("Simulation Input Parameters")
        main.geometry("400x500")

        rows = 20
        cols = 2

        main.rowconfigure(rows,weight=1)
        main.columnconfigure(cols,weight=1)

        nb = ttk.Notebook(main)
        nb.grid(row=1,column=0,columnspan=20,sticky="NESW")
        nb.enable_traversal()

        geometry_tab = ttk.Frame(nb)
        nb.add(geometry_tab,text="Geometry")

        lofting_tab = ttk.Frame(nb)
        nb.add(lofting_tab,text="Lofting")

        modeling_tab = ttk.Frame(nb)
        nb.add(modeling_tab,text="Modeling")

        meshing_tab = ttk.Frame(nb)
        nb.add(meshing_tab,text="Meshing")

        simulation_tab = ttk.Frame(nb)
        nb.add(simulation_tab,text="Simulation")

        #Options for Geometry Tab
        geom_label1 = ttk.Label(geometry_tab,text="Number of Sample Points")
        geom_label1.grid(row=1,column=0,sticky="W")

        geom_entry1 = ttk.Entry(geometry_tab)
        geom_entry1.insert(0,"50")
        geom_entry1.grid(row=1,column=1,sticky="E")

        geom_label2 = ttk.Label(geometry_tab,text="Use Distance Alignment")
        geom_label2.grid(row=2,column=0,sticky="W")

        geom_entry2 = tk.StringVar()
        geom_entry2.set("False")
        geom_entry2_options = ["False","True"]
        geom_entry2_drop = tk.OptionMenu(geometry_tab,geom_entry2,*geom_entry2_options)
        geom_entry2_drop.grid(row=2,column=1,sticky="E")

        # Options for Lofting Tab
        loft_label1 = ttk.Label(lofting_tab,text="U Knot Span Type")
        loft_label1.grid(row=1,column=0,sticky="W")

        loft_entry1 = tk.StringVar()
        loft_entry1.set("derivative")
        loft_entry1_options = ['derivative','equal','average']
        loft_entry1_drop = tk.OptionMenu(lofting_tab,loft_entry1,*loft_entry1_options)
        loft_entry1_drop.grid(row=1,column=1,sticky="E")

        loft_label2 = ttk.Label(lofting_tab,text="U Parametric Span Type")
        loft_label2.grid(row=2,column=0,sticky="W")

        loft_entry2 = tk.StringVar()
        loft_entry2.set("centripetal")
        loft_entry2_options = ['centripetal','chord','equal']
        loft_entry2_drop = tk.OptionMenu(lofting_tab,loft_entry2,*loft_entry2_options)
        loft_entry2_drop.grid(row=2,column=1,sticky="E")

        loft_label3 = ttk.Label(lofting_tab,text="U Degree")
        loft_label3.grid(row=3,column=0,sticky="W")

        loft_entry3 = ttk.Entry(lofting_tab)
        loft_entry3.insert(0,"2")
        loft_entry3.grid(row=3,column=1,sticky="E")

        loft_label4 = ttk.Label(lofting_tab,text="V Knot Span Type")
        loft_label4.grid(row=4,column=0,sticky="W")

        loft_entry4 = tk.StringVar()
        loft_entry4.set("average")
        loft_entry4_options = ['derivative','equal','average']
        loft_entry4_drop = tk.OptionMenu(lofting_tab,loft_entry4,*loft_entry4_options)
        loft_entry4_drop.grid(row=4,column=1,sticky="E")

        loft_label5 = ttk.Label(lofting_tab,text="V Parametric Span Type")
        loft_label5.grid(row=5,column=0,sticky="W")

        loft_entry5 = tk.StringVar()
        loft_entry5.set("chord")
        loft_entry5_options = ['centripetal','chord','equal']
        loft_entry5_drop = tk.OptionMenu(lofting_tab,loft_entry5,*loft_entry5_options)
        loft_entry5_drop.grid(row=5,column=1,sticky="E")

        loft_label6 = ttk.Label(lofting_tab,text="V Degree")
        loft_label6.grid(row=6,column=0,sticky="W")

        loft_entry6 = ttk.Entry(lofting_tab)
        loft_entry6.insert(0,"2")
        loft_entry6.grid(row=6,column=1,sticky="E")

        loft_label7 = ttk.Label(lofting_tab,text="Boundary Face Angle")
        loft_label7.grid(row=7,column=0,sticky="W")

        loft_entry7 = ttk.Entry(lofting_tab)
        loft_entry7.insert(0,"45")
        loft_entry7.grid(row=7,column=1,sticky="E")

        #Options for Modeling Tab
        model_label1 = ttk.Label(modeling_tab,text="Minimum Number of Face Cells")
        model_label1.grid(row=1,column=0,sticky="W")

        model_entry1 = ttk.Entry(modeling_tab)
        model_entry1.insert(0,"200")
        model_entry1.grid(row=1,column=1,sticky="E")

        model_label2 = ttk.Label(modeling_tab,text="hmin")
        model_label2.grid(row=2,column=0,sticky="W")

        model_entry2 = ttk.Entry(modeling_tab)
        model_entry2.insert(0,"0.02")
        model_entry2.grid(row=2,column=1,sticky="E")

        model_label3 = ttk.Label(modeling_tab,text="hmax")
        model_label3.grid(row=3,column=0,sticky="W")

        model_entry3 = ttk.Entry(modeling_tab)
        model_entry3.insert(0,"0.02")
        model_entry3.grid(row=3,column=1,sticky="E")

        model_label4 = ttk.Label(modeling_tab,text="Boundary Face Angle")
        model_label4.grid(row=4,column=0,sticky="W")

        model_entry4 = ttk.Entry(modeling_tab)
        model_entry4.insert(0,"45")
        model_entry4.grid(row=4,column=1,sticky="E")

        #Options for Meshing Tab

        mesh_label1 = ttk.Label(meshing_tab,text="Global Edge Size")
        mesh_label1.grid(row=1,column=0,sticky="W")

        mesh_entry1 = ttk.Entry(meshing_tab)
        mesh_entry1.insert(0,"0.01")
        mesh_entry1.grid(row=1,column=1,sticky="E")

        mesh_label2 = ttk.Label(meshing_tab,text="Surface Mesh Flag")
        mesh_label2.grid(row=2,column=0,sticky="W")

        mesh_entry2 = tk.StringVar()
        mesh_entry2.set("True")
        mesh_entry2_options = ["True","False"]
        mesh_entry2_drop = tk.OptionMenu(meshing_tab,mesh_entry2,*mesh_entry2_options)
        mesh_entry2_drop.grid(row=2,column=1,sticky="E")

        mesh_label3 = ttk.Label(meshing_tab,text="Volume Mesh Flag")
        mesh_label3.grid(row=3,column=0,sticky="W")

        mesh_entry3 = tk.StringVar()
        mesh_entry3.set("True")
        mesh_entry3_options = ["True","False"]
        mesh_entry3_drop = tk.OptionMenu(meshing_tab,mesh_entry3,*mesh_entry3_options)
        mesh_entry3_drop.grid(row=3,column=1,sticky="E")

        mesh_label4 = ttk.Label(meshing_tab,text="No Merge")
        mesh_label4.grid(row=4,column=0,sticky="W")

        mesh_entry4 = tk.StringVar()
        mesh_entry4.set("False")
        mesh_entry4_options = ["True","False"]
        mesh_entry4_drop = tk.OptionMenu(meshing_tab,mesh_entry4,*mesh_entry4_options)
        mesh_entry4_drop.grid(row=4,column=1,sticky="E")

        mesh_label5 = ttk.Label(meshing_tab,text="Optimization Level")
        mesh_label5.grid(row=5,column=0,sticky="W")

        mesh_entry5 = tk.StringVar()
        mesh_entry5.set("5")
        mesh_entry5_options = ["1","2","3","4","5"]
        mesh_entry5_drop = tk.OptionMenu(meshing_tab,mesh_entry5,*mesh_entry5_options)
        mesh_entry5_drop.grid(row=5,column=1,sticky="E")

        mesh_label6 = ttk.Label(meshing_tab,text="Global Edge Size")
        mesh_label6.grid(row=6,column=0,sticky="W")

        mesh_entry6 = ttk.Entry(meshing_tab)
        mesh_entry6.insert(0,"18.0")
        mesh_entry6.grid(row=6,column=1,sticky="E")

        #Options for Simulation Tab
        sim_label1 = ttk.Label(simulation_tab,text="SV Presolver Filename")
        sim_label1.grid(row=1,column=0,sticky="W")

        sim_entry1 = ttk.Entry(simulation_tab)
        sim_entry1.insert(0,"cco")
        sim_entry1.grid(row=1,column=1,sticky="E")

        sim_label2 = ttk.Label(simulation_tab,text="Fluid Density")
        sim_label2.grid(row=2,column=0,sticky="W")

        sim_entry2 = ttk.Entry(simulation_tab)
        sim_entry2.insert(0,"1.06")
        sim_entry2.grid(row=2,column=1,sticky="E")

        sim_label3 = ttk.Label(simulation_tab,text="Fluid Viscosity")
        sim_label3.grid(row=3,column=0,sticky="W")

        sim_entry3 = ttk.Entry(simulation_tab)
        sim_entry3.insert(0,"0.04")
        sim_entry3.grid(row=3,column=1,sticky="E")

        sim_label4 = ttk.Label(simulation_tab,text="Initial Pressure")
        sim_label4.grid(row=4,column=0,sticky="W")

        sim_entry4 = ttk.Entry(simulation_tab)
        sim_entry4.insert(0,"0")
        sim_entry4.grid(row=4,column=1,sticky="E")

        sim_label5 = ttk.Label(simulation_tab,text="Initial Velocity")
        sim_label5.grid(row=5,column=0,sticky="W")

        sim_entry5 = ttk.Entry(simulation_tab)
        sim_entry5.insert(0,"[0.0001,0.0001,0.0001]")
        sim_entry5.grid(row=5,column=1,sticky="E")

        sim_label6 = ttk.Label(simulation_tab,text="BCT Analytical Shape")
        sim_label6.grid(row=6,column=0,sticky="W")

        sim_entry6 = ttk.Entry(simulation_tab)
        sim_entry6.insert(0,"parabolic")
        sim_entry6.grid(row=6,column=1,sticky="E")

        sim_label7 = ttk.Label(simulation_tab,text="Inflow Filename")
        sim_label7.grid(row=7,column=0,sticky="W")

        sim_entry7 = ttk.Entry(simulation_tab)
        sim_entry7.insert(0,"inflow")
        sim_entry7.grid(row=7,column=1,sticky="E")

        sim_label8 = ttk.Label(simulation_tab,text="Period")
        sim_label8.grid(row=8,column=0,sticky="W")

        sim_entry8 = ttk.Entry(simulation_tab)
        sim_entry8.insert(0,"1")
        sim_entry8.grid(row=8,column=1,sticky="E")

        sim_label9 = ttk.Label(simulation_tab,text="BCT Point Number")
        sim_label9.grid(row=9,column=0,sticky="W")

        sim_entry9 = ttk.Entry(simulation_tab)
        sim_entry9.insert(0,"2")
        sim_entry9.grid(row=9,column=1,sticky="E")

        sim_label10 = ttk.Label(simulation_tab,text="Fourier Mode")
        sim_label10.grid(row=10,column=0,sticky="W")

        sim_entry10 = ttk.Entry(simulation_tab)
        sim_entry10.insert(0,"1")
        sim_entry10.grid(row=10,column=1,sticky="E")

        sim_label11 = ttk.Label(simulation_tab,text="Pressures")
        sim_label11.grid(row=11,column=0,sticky="W")

        sim_entry11 = ttk.Entry(simulation_tab)
        sim_entry11.insert(0,"0")
        sim_entry11.grid(row=11,column=1,sticky="E")

        sim_label12 = ttk.Label(simulation_tab,text="svsolver Filename")
        sim_label12.grid(row=12,column=0,sticky="W")

        sim_entry12 = ttk.Entry(simulation_tab)
        sim_entry12.insert(0,"solver")
        sim_entry12.grid(row=12,column=1,sticky="E")

        sim_label13 = ttk.Label(simulation_tab,text="Number of Timesteps")
        sim_label13.grid(row=13,column=0,sticky="W")

        sim_entry13 = ttk.Entry(simulation_tab)
        sim_entry13.insert(0,"200")
        sim_entry13.grid(row=13,column=1,sticky="E")

        sim_label14 = ttk.Label(simulation_tab,text="Timestep Size")
        sim_label14.grid(row=14,column=0,sticky="W")

        sim_entry14 = ttk.Entry(simulation_tab)
        sim_entry14.insert(0,"0.02")
        sim_entry14.grid(row=14,column=1,sticky="E")

        sim_label15 = ttk.Label(simulation_tab,text="Number of Timesteps Between Restarts")
        sim_label15.grid(row=15,column=0,sticky="W")

        sim_entry15 = ttk.Entry(simulation_tab)
        sim_entry15.insert(0,"50")
        sim_entry15.grid(row=15,column=1,sticky="E")

        sim_label16 = ttk.Label(simulation_tab,text="Number of Force Surfaces")
        sim_label16.grid(row=16,column=0,sticky="W")

        sim_entry16 = ttk.Entry(simulation_tab)
        sim_entry16.insert(0,"1")
        sim_entry16.grid(row=16,column=1,sticky="E")

        sim_label17 = ttk.Label(simulation_tab,text="ID of Force Surface Calculation")
        sim_label17.grid(row=17,column=0,sticky="W")

        sim_entry17 = ttk.Entry(simulation_tab)
        sim_entry17.insert(0,"1")
        sim_entry17.grid(row=17,column=1,sticky="E")

        sim_label18 = ttk.Label(simulation_tab,text="Force Calculation Method")
        sim_label18.grid(row=18,column=0,sticky="W")

        sim_entry18 = ttk.Entry(simulation_tab)
        sim_entry18.insert(0,"Velocity Based")
        sim_entry18.grid(row=18,column=1,sticky="E")
        """
                          svpre_name='cco',fluid_density=1.06,
                          fluid_viscosity=0.04, initial_pressure=0,
                          initial_velocity=[0.0001,0.0001,0.0001],
                          bct_analytical_shape='parabolic',
                          inflow_file='inflow',
                          period=1,bct_point_number=2,
                          fourier_mode=1,pressures=0,
                          svsolver='solver',number_timesteps=200,
                          timestep_size=0.02,number_restarts=50,
                          number_force_surfaces=1,
                          surface_id_force_calc=1,
                          force_calc_method='Velocity Based',
                          print_avg_solution=True,
                          print_error_indicators=False,
                          varying_time_from_file=True,
                          step_construction='0 1 0 1',
                          pressure_coupling='Implicit',
                          backflow_stabilization=0.2,
                          residual_control=True,
                          residual_criteria=0.01,
                          minimum_req_iter=2,
                          svLS_type='NS',num_krylov=100,
                          num_solves_per_left=1,
                          tolerance_momentum=0.05,
                          tolerance_continuity=0.4,
                          tolerance_svLS_NS=0.4,
                          max_iter_NS=2,max_iter_momentum=4,
                          max_iter_continuity=400,
                          time_integration_rule='Second Order',
                          time_integration_rho=0.5,
                          flow_advection_form='Convective',
                          quadrature_interior=2,
                          quadrature_boundary=3,procs=24,
                          svpost='cco',start=0,vtu=True,
                          vtp=False,vtkcombo=False,
                          all_arg=True,wss=False,
                          sim_units_mm=False,
                          sim_units_cm=True,
                          global_edge_size=0.01,
                          distal_resistance=0
    """



# Remeshing utility based on MMG executables

import os
import platform
import subprocess
import pyvista as pv
import io
import sys
filepath = os.path.abspath(__file__)
dirpath  = os.path.dirname(filepath)

def remesh_surface(pv_polydata_object,auto=True,hausd=0.01,verbosity=1):
    _mesh_ = pv.PolyData(pv_polydata_object.points,pv_polydata_object.faces)
    pv.save_meshio("tmp.mesh",_mesh_)
    if platform.system() == 'Windows':
        _EXE_ = dirpath+os.sep+"Windows"+os.sep+"mmgs_O3.exe"
    elif platform.system() == "Linux":
        _EXE_ = dirpath+os.sep+"Linux"+os.sep+"mmgs_O3"
    elif platform.system() == "Darwin":
        _EXE_ = dirpath+os.sep+"Mac"+os.sep+"mmgs_O3"
    devnull = open(os.devnull, 'w')
    if verbosity == 0:
        subprocess.check_call([_EXE_,"tmp.mesh","-hausd",str(hausd),"-v",str(verbosity)],stdout=devnull,stderr=devnull)
    else:
        subprocess.check_call([_EXE_,"tmp.mesh","-hausd",str(hausd),"-v",str(verbosity)])
    clean_medit("tmp.o.mesh")
    remeshed = pv.read("tmp.o.mesh")
    remeshed_surface = remeshed.extract_surface()
    remeshed_surface.triangulate(inplace=True)
    remeshed_surface.clear_data()
    os.remove("tmp.mesh")
    os.remove("tmp.o.sol")
    os.remove("tmp.o.mesh")
    return remeshed_surface

def clean_medit(filename):
    file = open(filename)
    lines = file.readlines()
    file.close()
    keywords_index = []
    for i,s in enumerate(lines):
        if s[0].isnumeric():
            pass
        elif s[0] == '-':
            pass
        elif s[0] == '\n':
            pass
        elif s[0] == '\n':
            pass
        else:
            keywords_index.append(i)
    new_file = open(filename,'w+')
    new_lines = []
    for i,o in enumerate(keywords_index):
        if lines[o] == 'RequiredVertices\n':
            pass
        elif lines[o] == 'Ridges\n':
            pass
        elif lines[o] == 'Tangents\n':
            pass
        elif lines[o] == 'TangentAtVertices\n':
            pass
        else:
            if o == keywords_index[-1]:
                new_lines.append(lines[o])
            else:
                new_lines.extend(lines[o:keywords_index[i+1]])
    new_file.writelines(new_lines)
    new_file.close()



import numpy as np
from .basis import *
from .calculate_length import *
from .update import *
from .calculate_radii import *

def add_geodesic_path(data,path,lengths,parent_vessel_idx,Qterm,gamma,nu,
                      Pperm, Pterm, sub_division_index,sub_division_map):
    # Run if the added path is the root vessel
    if np.all(data[0,:]==0):
        for i in range(len(path)-1):
            data[i,0:3] = path[i]
            data[i,3:6] = path[i+1]
            basis(data,i)
            data[i,15] = i+1
            data[i,16] = -1
            data[i,17] = i - 1
            data[i,18] = i
            data[i,19] = i+1
            data[i,20] = lengths[i] #length(data,i)
            data[i,22] = Qterm
            if idx < len(path) - 2:
                data[i,23] = 1.0
            else:
                data[i,23] = 0.0
            data[i,24] = 0
            data[i,25] = (8*nu/np.pi)*np.sum(lengths[i:])
            data[i,26] = 0
            data[i,27] = np.sum(lengths[i:])
            data[i,28] = 1.0
            data[i,29] = -1
            data[i,-1] = 0
            data[i,21] = ((data[0,25]*data[0,22])/(Pperm-Pterm)) ** (1/4)
            sub_division = [-1]
            sub_division_index.append(len(sub_division_map))
            sub_division.extend(list(range(i+1,data.shape[0])))
            sub_division_map.extend(sub_division)
    else:
        ## This is going to be a bit complicated
        ## TODO needs an update strategy for updaing index 28 of vessels
        ## the subvessels are of the same depth which needs to be considered
        ## when making a strategy.

        ############################
        # Add Geodesic Terminal Vessel
        ############################
        count = int(max(data[:,19]))
        new_vessels = []
        for idx in range(len(path)-1):
            i = idx + count
            data[i,0:3] = path[idx]
            data[i,3:6] = path[idx+1]
            basis(data,i)
            if idx < len(path) - 2:
                data[i,15] = i+1
            else:
                data[i,15] = -1
            data[i,16] = -1
            if idx == 0:
                data[i,17] = data[parent_vessel_idx,-1]
            else:
                data[i,17] = i - 1
            if idx == 0:
                bif_node = i+1
                bif_vessel = i
            data[i,18] = i+1
            data[i,19] = i+2
            data[i,20] = lengths[idx] #length(data,i)
            data[i,22] = Qterm
            if idx < len(path) - 2:
                data[i,23] = 1.0
            else:
                data[i,23] = 0.0
            data[i,24] = 0
            data[i,25] = (8*nu/np.pi)*np.sum(lengths[idx:])
            data[i,26] = data[parent_vessel_idx,26]+1
            if idx < len(path) - 2:
                data[i,27] = np.sum(lengths[idx+1:])
            else:
                data[i,27] = 0.0
            data[i,28] = 1.0
            data[i,29] = -1
            data[i,-1] = i
            sub_division = [-1]
            #sub_division_index.append(sub_division_map)
            new_vessels.append(i)
            if idx < len(path) - 2:
                sub_division.extend(list(range(i+1,data.shape[0]-1)))
            sub_division_map.extend(sub_division)
        ########################
        # Add Sister vessel
        ########################
        data[-1,0:3] = path[0]
        data[-1,3:6] = data[parent_vessel_idx,3:6]
        basis(data,data.shape[0]-1)
        data[-1,15]  = data[parent_vessel_idx,15]
        data[-1,16]  = data[parent_vessel_idx,16]
        data[-1,17]  = data[parent_vessel_idx,-1]
        data[-1,18]  = bif_node
        data[-1,19]  = data[parent_vessel_idx,19]
        length(data,data.shape[0]-1)
        data[-1,22]  = data[parent_vessel_idx,22]
        data[-1,23]  = data[parent_vessel_idx,23]
        data[-1,24]  = data[parent_vessel_idx,24]
        data[-1,25]  = data[parent_vessel_idx,25] - data[parent_vessel_idx,20]*(8*nu/np.pi) + data[-1,20]*(8*nu/np.pi)
        data[-1,26]  = data[parent_vessel_idx,26]+1
        data[-1,27]  = data[parent_vessel_idx,27]
        data[-1,28]  = data[parent_vessel_idx,28]
        data[-1,29]  = -1
        data[-1,-1]  = data.shape[0]-1
        print('PARENT 28: {}'.format(data[parent_vessel_idx,28]))
        ########################
        # Replace children parent id with new id
        ########################
        left = int(data[-1,15])
        right = int(data[-1,16])
        if left != -1:
            data[left,17] = data[-1,-1]
        if right != -1:
            data[right,17] = data[-1,-1]
        new_vessels.append(data.shape[0]-1)
        ########################
        # Update parent vessel
        ########################
        data[parent_vessel_idx,3:6] = path[0]
        basis(data,parent_vessel_idx)
        data[parent_vessel_idx,15]  = bif_vessel
        data[parent_vessel_idx,16]  = data.shape[0]-1
        data[parent_vessel_idx,19]  = bif_node
        length(data,parent_vessel_idx)
        #data[parent_vessel_idx,22]  += Qterm
        sub_division_index = np.argwhere(np.array(sub_division_map)==-1).flatten()
        left = int(data[parent_vessel_idx,15])
        right = int(data[parent_vessel_idx,16])
        LR = ((data[left,22]*data[left,25])/(data[right,22]*data[right,25])) **(1/4)
        lbif = (1 + LR ** (-gamma)) ** (-1/gamma)
        rbif = (1 + LR ** (gamma)) ** (-1/gamma)
        sister_sub_division = [-1]
        if data[-1,15] > 0 or data[-1,16] > 0:
            start = sub_division_index[parent_vessel_idx]
            if sub_division_index[parent_vessel_idx] == sub_division_index[-1]:
                end = None
            else:
                end = sub_division_index[parent_vessel_idx+1]
            k = sub_division_map[start+1:end]
            #data[k,28] = (data[k,28]/data[parent_vessel_idx,28])*rbif
            print(k)
            data[k,26] += 1
            sister_sub_division.extend(k)
        data[parent_vessel_idx,23] = lbif
        data[parent_vessel_idx,24] = rbif
        #data[parent_vessel_idx,25] = (8*nu/np.pi)*data[parent_vessel_idx] + ((lbif ** 4 /data[left,25]) + \
        #                                                  (rbif ** 4 /data[right,25])) ** -1
        data[parent_vessel_idx,27] = lbif**2 * (data[left,20] + data[left,27]) +\
                                     rbif**2 * (data[right,20] + data[right,27])
        #data[parent_vessel_idx,28] = 1
        sub_division_map.extend(sister_sub_division)
        parent_sub_idx = sub_division_index[parent_vessel_idx]
        sub_division_map[parent_sub_idx+1:parent_sub_idx+1] = new_vessels
        sub_division_index = np.argwhere(np.array(sub_division_map)==-1).flatten()
        main_idx  = []
        alt_idx   = []
        while parent_vessel_idx >= 0:
            left = int(data[parent_vessel_idx,15])
            right = int(data[parent_vessel_idx,16])
            data[parent_vessel_idx,22] += Qterm
            main_idx.append(parent_vessel_idx)
            #if left in main_idx:
            #    alt_idx.append(right)
            #else:
            #    alt_idx.append(left)
            if left > 0 and right > 0:
                LR = ((data[left,22]*data[left,25])/(data[right,22]*data[right,25])) ** (1/4)
                lbif = (1 + LR ** (-gamma)) ** (-1/gamma)
                rbif = (1 + LR ** (gamma)) ** (-1/gamma)
            elif left > 0 and right < 0:
                lbif  = 1
                rbif  = 0
            elif right > 0 and left <  0:
                lbif = 0
                rbif = 1
            else:
                lbif = 0
                rbif = 0
            if left > 0 and right > 0:
                data[parent_vessel_idx,25] = (8*nu/np.pi) * data[parent_vessel_idx,20] +\
                                             ((lbif**4/ data[left,25]) +
                                             (rbif**4/ data[right,25])) ** -1
            elif left > 0 and right < 0:
                data[parent_vessel_idx,25] = (8*nu/np.pi) * data[parent_vessel_idx,20] +\
                                             data[left,25]
            elif right > 0 and left < 0:
                data[parent_vessel_idx,25] = (8*nu/np.pi) * data[parent_vessel_idx,20] +\
                                             data[right,25]
            data[parent_vessel_idx,23] = lbif
            data[parent_vessel_idx,24] = rbif
            data[parent_vessel_idx,27] = lbif ** 2 * (data[left,20] + data[left,27]) +\
                                         rbif ** 2 * (data[right,20] + data[right,27])
            parent_sub_index = sub_division_index[parent_vessel_idx]
            sub_division_map[parent_sub_index+1:parent_sub_index+1] = new_vessels
            #print(len(sub_division_map))
            sub_division_index = np.argwhere(np.array(sub_division_map)==-1).flatten()
            parent_vessel_idx = int(data[parent_vessel_idx,17])
            left = int(data[parent_vessel_idx,15])
            right = int(data[parent_vessel_idx,16])
            if parent_vessel_idx > 0:
                if left in main_idx:
                    alt_idx.append(right)
                else:
                    alt_idx.append(left)
            else:
                alt_idx.append(-1)
        #print(main_idx)
        #print(alt_idx)
        data[0,21] = ((data[0,25]*data[0,22])/(Pperm-Pterm)) ** (1/4)
        main_idx = main_idx[::-1]
        alt_idx = alt_idx[::-1]
        #print('main idx: {}'.format(len(main_idx)))
        #print('alt  idx: {}'.format(len(alt_idx)))
        #main_idx.pop(0)
        #alt_idx.pop(0)
        new_main_scale = 1
        new_alt_scale = 1

        if int(data[main_idx[0],15]) in main_idx:
            new_main_scale *= data[main_idx[0],23]
            #if alt_idx[0] > 0:
            new_alt_scale *= data[main_idx[0],24]

        if int(data[main_idx[0],16]) in main_idx:
            new_main_scale *= data[main_idx[0],24]
            #if alt_idx[0] > 0:
            new_alt_scale = data[main_idx[0],23]
        main_idx.pop(0)
        alt_idx.pop(0)
        print('main_idx: {}'.format(main_idx))
        print('alt_idx:  {}'.format(alt_idx))
        if len(main_idx) == 0:
            print('main idx length = 0')
        for i in range(len(main_idx)):
            print('main: {}, scale: {}'.format(main_idx[i],new_main_scale))
            print('alt: {},  scale: {}'.format(alt_idx[i],new_alt_scale))
            data[main_idx[i],28] = new_main_scale
            start = sub_division_index[main_idx[i]]
            if main_idx[i] == sub_division_index[-1]:
                end = None
            else:
                end = sub_division_index[main_idx[i]+1]
            if alt_idx[i] > -1:
                alt_start = sub_division_index[alt_idx[i]]
                if sub_division_index[alt_idx[i]] == sub_division_index[-1]:
                    alt_end = None
                else:
                    alt_end = sub_division_index[alt_idx[i]+1]
                tmp = sub_division_map[alt_start+1:alt_end]
                data[tmp,28] = (data[tmp,28]/data[alt_idx[i],28])*(new_alt_scale)
                data[alt_idx[i],28] = new_alt_scale
            """
            if len(sub_division_map[start+1:end]) > 0 and main_idx[i] > 0:
                if alt_idx[i] > -1:
                    alt_start = sub_division_index[alt_idx[i]]
                    if sub_division_index[alt_idx[i]] == sub_division_index[-1]:
                        alt_end = None
                    else:
                        alt_end = sub_division_index[alt_idx[i]+1]
                    tmp = sub_division_map[alt_start+1:alt_end]
                    data[tmp,28] = (data[tmp,28]/data[alt_idx[i],28])*new_alt_scale
            """
            #if alt_idx[i] > 0:
            #    data[alt_idx[i],28] = new_alt_scale
            #    #data[alt_idx[i],21] = new_alt_scale*data[0,21]
            if int(data[main_idx[i],15]) in main_idx:
                #if alt_idx[0] > 0:
                new_alt_scale = new_main_scale*data[main_idx[i],24]
                new_main_scale *= data[main_idx[i],23]
                #else:
                #    new_alt_scale *= data[0,23]
            if int(data[main_idx[i],16]) in main_idx:
                #if alt_idx[0] > 0:
                new_alt_scale = new_main_scale*data[main_idx[i],23]
                new_main_scale *= data[main_idx[i],24]
                #else:
                #    new_alt_scale *= data[0,24]
        ########################
        # Update geodesic vessel
        ########################
        start = sub_division_index[bif_vessel]
        if sub_division_index[bif_vessel] == sub_division_index[-1]:
            end = None
        else:
            end = sub_division_index[bif_vessel+1]
        tmp = sub_division_map[start+1:end]
        #print(tmp)
        #print(new_main_scale)
        data[tmp,28] = new_main_scale
        data[bif_vessel,28] = new_main_scale
        print('SISTER ALT SCALE: {}'.format(new_alt_scale))
        #data[data.shape[0]-1,28] = data[main_idx[-1],24]
        #rbif = data[main_idx[-1],24]
        #if len(sub_division_map[start+1:end]) > 0:
        #    if (data.shape[0]-1) > -1:
        alt_start = sub_division_index[data.shape[0] - 1]
        if sub_division_index[data.shape[0]-1] == sub_division_index[-1]:
            alt_end = None
        else:
            alt_end = sub_division_index[data.shape[0]-1+1]
        tmp = sub_division_map[alt_start+1:alt_end]
        data[tmp,28] = (data[tmp,28]/data[data.shape[0]-1,28])*(new_alt_scale)#data[data.shape[0]-1,28])*new_alt_scale
        data[data.shape[0]-1,28] = new_alt_scale
        data[:,21] = data[:,28]*data[0,21]
    return data,sub_division_map,sub_division_index



"""
Function for installing extra external packages from within the
sv python interpreter
"""

install_function="""
from pip._internal import main as pipmain
def install(packagename,version=None):
    try:
        if version is None:
            pipmain(['install',packagename])
        else:
            pipmain(['install',packagename+'=='+version])
    except:
        print('package installation failed')
"""

uninstall_function="""
def uninstall(packagename):
    try:
        pipmain(['uninstall',packagename])
    except:
        print('package uninstallation failed')
"""



view_plots="""import numpy as np
import matplotlib.pyplot as plt
data = np.load("solver_0d_branch_results.npy",allow_pickle=True).item()
vessels = []
fig_flow = plt.figure()
ax_flow = fig_flow.add_subplot()
fig_pressure = plt.figure()
ax_pressure = fig_pressure.add_subplot()
fig_flow_outlets = plt.figure()
ax_flow_outlets = fig_flow_outlets.add_subplot()
fig_pressure_outlets = plt.figure()
ax_pressure_outlets = fig_pressure_outlets.add_subplot()
for vessel in data["flow"]:
    vessels.append(vessel)
for vessel in vessels:
    ax_flow.plot(data["time"],data["flow"][vessel][0],label="vessel_"+str(vessel))
    ax_pressure.plot(data["time"],data["pressure"][vessel][0]/1333.22,label="vessel_"+str(vessel))
import json
info = json.load(open("solver_0d.in"))
all_inlets = []
all_outlets = []
for i in info["junctions"]:
    for j in i["inlet_vessels"]:
        all_inlets.append(j)
    for j in i["outlet_vessels"]:
        all_outlets.append(j)
all_inlets = set(all_inlets)
all_outlets = set(all_outlets)
true_outlets = list(all_outlets.difference(all_inlets))
true_inlets = list(all_inlets.difference(all_outlets))
for vessel in true_outlets:
    ax_flow_outlets.plot(data["time"],data["flow"][vessel][-1],label="vessel_"+str(vessel))
    ax_pressure_outlets.plot(data["time"],data["pressure"][vessel][-1]/1333.22,label="vessel_"+str(vessel))
ax_flow.set_xlabel("Time (sec)")
ax_flow.set_ylabel("Flow (mL/s)")
ax_pressure.set_xlabel("Time (sec)")
ax_pressure.set_ylabel("Pressure (mmHg)")
ax_flow_outlets.set_xlabel("Time (sec)")
ax_pressure_outlets.set_xlabel("Time (sec)")
ax_flow_outlets.set_ylabel("Flow (mL/s)")
ax_flow_outlets.set_title("Outlets Only")
ax_pressure_outlets.set_ylabel("Pressure (mmHg)")
ax_pressure_outlets.set_title("Outlets Only")
#ax_pressure_outlets.set_ylim([-10*np.finfo(float).eps,10*np.finfo(float).eps])
plt.show()

"""



import numpy as np
import numba as nb
from functools import partial

@nb.jit(nopython=True,cache=True)
def fast_local_function(data,mesh,terminal,edge,gamma,nu,Qterm,Pperm,Pterm):
    ####################
    reduced_resistances = []
    reduced_lengths     = []
    bifurcations        = []
    flows               = [] # should only be changing path (not alt paths)
    main_idx            = [] # index 0: data[:,15]; index 1: data[:,16]
    alt_idx             = []
    new_scale           = [] # compound bifurcations
    alt_scale           = []
    ####################
    total_flow          = data[edge,22]+Qterm*2
    upstream            = data[edge,0:3]
    downstream          = data[edge,3:6]
    upstream_length     = np.sqrt(np.sum(np.square(mesh-upstream),axis=1))
    downstream_length   = np.sqrt(np.sum(np.square(mesh-downstream),axis=1))
    terminal_length     = np.sqrt(np.sum(np.square(mesh-terminal),axis=1))
    # Begin calculating reduced resistance and bifurcation ratios
    R_terminal          = (8*nu/np.pi)*terminal_length
    R_terminal_sister   = (8*nu/np.pi)*downstream_length + (data[edge,25]-(8*nu/np.pi)*data[edge,20])
    f_terminal          = (1+((data[edge,22]*R_terminal_sister)/(Qterm*R_terminal))**(gamma/4)) ** (-1/gamma)
    f_terminal_sister   = (1+((data[edge,22]*R_terminal_sister)/(Qterm*R_terminal))**(-gamma/4)) ** (-1/gamma)
    R_0                 = (8*nu/np.pi)*upstream_length + ((f_terminal**4)/R_terminal +
                          (f_terminal_sister**4)/R_terminal_sister)**-1
    L_0                 = f_terminal ** 2 * terminal_length + f_terminal_sister ** 2 * (downstream_length + data[edge,27])
    reduced_resistances.append(R_0)
    reduced_lengths.append(L_0)
    flows.append(data[edge,22]+Qterm)
    main_idx.append(edge)
    bifurcations.append(f_terminal_sister) # 0 -> 15; 1-> 16
    bifurcations.append(f_terminal)
    if edge == 0:
        # on the terminating round the new terminal is always 16 and the existing is 15
        #bifurcations.append([f_terminal_sister,f_terminal])
        L_0 = upstream_length + L_0
        R_0 = (((data[edge,22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
        alt_idx.append(-1)
        new_scale.append(np.ones(f_terminal.shape))
        alt_scale.append(np.ones(f_terminal.shape))
        return (R_0,L_0,f_terminal,f_terminal_sister,None,mesh,reduced_resistances,
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
                R_terminal_sister)
    previous = edge
    edge = int(data[edge,17])
    if int(data[edge,15]) == previous:
        alt = int(data[edge,16])
        alt_idx.append(alt)
        position = 0
        #bifurcations.append([f_terminal,f_terminal_sister])
    else:
        alt = int(data[edge,15])
        alt_idx.append(alt)
        position = 1
        #bifurcations.append([f_terminal_sister,f_terminal])
    f_changed = (1+((data[alt,22]*data[alt,25])/((data[previous,22]+Qterm)*R_0)) ** (gamma/4)) ** (-1/gamma)
    f_stagnant = (1+((data[alt,22]*data[alt,25])/((data[previous,22]+Qterm)*R_0)) ** (-gamma/4)) ** (-1/gamma)
    f_terminal = f_terminal*f_changed
    f_terminal_sister = f_terminal_sister*f_changed
    f_parent = f_changed
    R_0 = (8*nu/np.pi)*data[edge,20] + ((f_changed ** 4) /R_0 + (f_stagnant ** 4) /data[alt,25]) ** -1
    L_0 = f_changed ** 2 *(upstream_length+L_0) + f_stagnant ** 2 * (data[alt,20] + data[alt,27])
    reduced_resistances.append(R_0)
    reduced_lengths.append(L_0)
    flows.append(data[edge,22]+Qterm)
    main_idx.append(edge)
    new_scale.append(f_changed)
    alt_scale.append(f_stagnant)
    if position == 0:
        bifurcations.append(f_changed)
        bifurcations.append(f_stagnant)
    else:
        bifurcations.append(f_stagnant)
        bifurcations.append(f_changed)
    if edge == 0:
        L_0 = data[edge,20]+L_0
        R_0 = (((data[edge,22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
        return (R_0,L_0,f_terminal,f_terminal_sister,f_parent,mesh,reduced_resistances,
                reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
                R_terminal_sister)
    previous = edge
    edge = int(data[edge,17])
    while edge >= 0:
        if int(data[edge, 15]) == previous:
            alt = int(data[edge,16])
            alt_idx.append(alt)
            position = 0
        else:
            alt = int(data[edge,15])
            alt_idx.append(alt)
            position = 1
        f_changed = (1+((data[alt, 22]*data[alt,25])/((Qterm+data[previous,22])*R_0)) ** (gamma/4)) ** (-1/gamma)
        f_stagnant = (1+((data[alt, 22]*data[alt,25])/((Qterm+data[previous,22])*R_0)) ** (-gamma/4)) ** (-1/gamma)
        f_terminal = f_terminal*f_changed
        f_terminal_sister = f_terminal_sister*f_changed
        f_parent = f_parent*f_changed
        R_0 = (8*nu/np.pi)*data[edge,20] + ((f_changed ** 4 / R_0) + (f_stagnant ** 4 / data[alt, 25])) ** -1
        L_0 = f_changed ** 2 * (data[previous, 20] + L_0) + f_stagnant ** 2 *(data[alt, 20] + data[alt, 27])
        reduced_resistances.append(R_0)
        reduced_lengths.append(L_0)
        flows.append(data[edge,22]+Qterm)
        main_idx.append(edge)
        if edge >= 0:
            for j in range(len(new_scale)):
                new_scale[j] = new_scale[j]*f_changed
                alt_scale[j] = alt_scale[j]*f_changed
            new_scale.append(f_changed)
            #alt_scale = [bif*f_changed for bif in alt_scale]
            alt_scale.append(f_stagnant)
        #else:
        #    new_scale.append(np.ones(f_changed.shape))
        #    alt_scale.append(np.ones(f_changed.shape))
        if position == 0:
            bifurcations.append(f_changed)
            bifurcations.append(f_stagnant)
        else:
            bifurcations.append(f_stagnant)
            bifurcations.append(f_changed)
        previous = edge
        edge = int(data[edge,17])
    L_0 = data[0,20] + L_0
    R_0 = (((data[0, 22]+Qterm)*R_0)/(Pperm-Pterm))**(1/4)
    return (R_0,L_0,f_terminal,f_terminal_sister,f_parent,mesh,reduced_resistances,
            reduced_lengths,bifurcations,flows,main_idx,alt_idx,new_scale,alt_scale,R_terminal,
            R_terminal_sister)



"Locate Solver Files"
import os
import platform

def locate_0d_solver(windows_drive="C",linux_drive=os.sep):
    path = None
    if platform.system() == "Windows":
        for root, subdirs, files in os.walk(windows_drive+":"+os.sep):
            for d in subdirs:
                if d == 'svZeroDSolver' and '.git' not in root:
                    path = root
                    escape = True
                    break
                else:
                    escape = False
            if escape:
                break
    else:
        for root, subdirs, files in os.walk(linux_drive):
            for d in subdirs:
                if d == 'svZeroDSolver' and '.git' not in root:
                    path = root
                    escape = True
                    break
                else:
                    escape = False
            if escape:
                break
    return path

def locate_1d_solver(windows_drive="C",linux_drive=os.sep):
    path = None
    if platform.system() == "Windows":
        for root, subdirs, files in os.walk(windows_drive+":"+os.sep):
            for f in files:
                if f == 'svOneDSolver.exe':
                    path = root+os.sep+d+os.sep+f
                    escape = True
                    break
                else:
                    escape = False
            if escape:
                break
    else:
        for root, subdirs, files in os.walk(linux_drive):
            for f in files:
                if f == 'OneDSolver':
                    path = root+os.sep+d+os.sep+f
                    escape = True
                    break
                else:
                    escape = False
            if escape:
                break
    return path



# Reconstructed Implicit Surfaces
import svcco
import pyvista as pv
from time import perf_counter
import os
import numpy as np
import matplotlib.pyplot as plt
import tqdm
import tetgen

###################################
# Setting up surface reconstruction
###################################

cu = pv.Cube(x_length=3.72,y_length=3.72,z_length=3.72).triangulate().subdivide(3)
cube = svcco.surface()
cube.set_data(cu.points,cu.point_normals)
cube.solve()
cube.build(q=2,resolution=30)
cube.pv_polydata_surf.save('cube.vtp')

heart = svcco.surface()
heart_points = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_points_unique.csv',delimiter=',')
heart_normals = np.genfromtxt('D:\\svcco\\svcco\\implicit\\tests\\heart_normals_unique.csv',delimiter=',')
heart.set_data(heart_points,heart_normals)
heart.solve()
heart.build(q=4,resolution=120,k=2,buffer=5)
heart.pv_polydata_surf.save('heart.vtp')

disk = pv.Disc(r_res=20,c_res=100)
cyl  = disk.extrude([0,0,1],capping=True).triangulate()
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl)
cyl  = cyl.subdivide(2)
cyl  = svcco.utils.remeshing.remesh.remesh_surface(cyl,hausd=0.005)
cyl  = cyl.compute_normals(auto_orient_normals=True,feature_angle=90)
cylinder = svcco.surface()
cylinder.set_data(cyl.points,cyl.point_normals)
cylinder.solve()
cylinder.build(q=10,k=2,resolution=25,method='flying_edges')
cylinder.pv_polydata_surf.save('cylinder.vtp')


left_gyrus   = "D:\\Tree\\Tree_8-0\\brain_testing\\FJ3801_BP58201_FMA72658_Left inferior frontal gyrus.obj"
gyrus = svcco.surface()
gyrus.load(left_gyrus)
gyrus.solve()
gyrus.build(q=4,resolution=40,buffer=5)


refined_cube = cube.pv_polydata.subdivide(2)
tet = tetgen.TetGen(refined_cube)
total_patches = len(cube.patches)
surface_pts = cube.pv_polydata.points
surface_pts = [pt.tolist() for pt in surface_pts]
values = []
tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5)
mesh = tet.grid
pts = mesh.points
for pt in pts:
    data = pt.tolist()
    data.append(total_patches//4)
    values.append(cube.DD[0](data)[0])
values = np.array(values)
mesh['implicit_data'] = values
mesh.save('cube.vtu')

refined_cyl = cylinder.pv_polydata.subdivide(2)
refined_cyl = svcco.utils.remeshing.remesh.remesh_surface(refined_cyl,hausd=0.005)
tet = tetgen.TetGen(refined_cyl)
tet.make_manifold(verbose=True)
total_patches = len(cylinder.patches)
surface_pts = cylinder.pv_polydata.points
surface_pts = [pt.tolist() for pt in surface_pts]
values = []
tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5)
mesh = tet.grid
pts = mesh.points
for pt in pts:
    data = pt.tolist()
    data.append(total_patches//4)
    values.append(cylinder.DD[0](data)[0])

values = np.array(values)
mesh['implicit_data'] = values
mesh.save('cylinder.vtu')

refined_heart = heart.pv_polydata
refined_heart = svcco.utils.remeshing.remesh.remesh_surface(refined_heart,hausd=0.005)
tet = tetgen.TetGen(refined_heart)
tet.make_manifold(verbose=True)
total_patches = len(heart.patches)
surface_pts = heart.pv_polydata.points
surface_pts = [pt.tolist() for pt in surface_pts]
values = []
tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5)
mesh = tet.grid
pts = mesh.points
for pt in pts:
    data = pt.tolist()
    data.append(total_patches//4)
    values.append(heart.DD[0](data)[0])

values = np.array(values)
mesh['implicit_data'] = values
mesh.save('cylinder.vtu')

refined_gyrus = gyrus.pv_polydata
refined_gyrus = svcco.utils.remeshing.remesh.remesh_surface(refined_gyrus,hausd=0.005)
tet = tetgen.TetGen(refined_gyrus)
tet.make_manifold(verbose=True)
total_patches = len(gyrus.patches)
surface_pts = gyrus.pv_polydata.points
surface_pts = [pt.tolist() for pt in surface_pts]
values = []
tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5)
mesh = tet.grid
pts = mesh.points
for pt in pts:
    data = pt.tolist()
    data.append(total_patches//4)
    values.append(gyrus.DD[0](data)[0])

values = np.array(values)
mesh['implicit_data'] = values
mesh.save('gyrus.vtu')
"""

"""

# Convex shapes
"""
sphere = pv.Sphere().triangulate().subdivide(2)

surf = svcco.surface()
surf.set_data(sphere.points,normals=sphere.point_normals)
surf.solve()
surf.build()

refined = sphere.subdivide(2)
tet = tetgen.TetGen(refined)
total_patches = len(surf.patches)
values = []
tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5)
mesh = tet.grid
centers = mesh.cell_centers()

for i in tqdm.tqdm(range(centers.n_cells)):
    data = centers.cell_points(i).flatten().tolist()
    data.append(total_patches)
    values.append(surf.DD[0](data))

values = np.array(values)
mesh.cell_data['implicit_data'] = values
mesh.save('sphere.vtu')
"""

disc = pv.Disc(r_res=10,c_res=20)
cyl  = disc.extrude([0,0,1],capping=True)
cyl.triangulate(inplace=True)
cyl = svcco.utils.remeshing.remesh.remesh_surface(cyl)

surf = svcco.surface()
surf.set_data(cyl.points,normals=cyl.point_normals)
surf.solve()
surf.build()

refined = cyl.subdivide(2)
tet = tetgen.TetGen(refined)
total_patches = len(surf.patches)
values = []
tet.tetrahedralize(order=1,mindihedral=20,minratio=1.5)
mesh = tet.grid
centers = mesh.cell_centers()

for i in tqdm.tqdm(range(centers.n_cells)):
    data = centers.cell_points(i).flatten().tolist()
    data.append(total_patches)
    values.append(surf.DD[0](data))

values = np.array(values)
mesh.cell_data['implicit_data'] = values
mesh.save('cylinder.vtu')
